@@ -1,5 +1,4 @@
-# -*- coding: utf-8 -*-
-# Copyright 2007-2018 Red Hat, Inc. and others.
+# Copyright 2007-2019 Red Hat, Inc. and others.
 #
 # This file is part of Bodhi.
 #
@@ -18,33 +17,41 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 """Define tools for interacting with the build system and a fake build system for development."""
 
+from functools import wraps
 from threading import Lock
 import logging
+import os
 import time
-from functools import wraps
+import typing
 
+import backoff
 import koji
 
 
+if typing.TYPE_CHECKING:  # pragma: no cover
+    from bodhi.server.config import BodhiConfig  # noqa: 401
+
+
 log = logging.getLogger('bodhi')
 _buildsystem = None
+_buildsystem_login_lock = Lock()
 # URL of the koji hub
 _koji_hub = None
 
 
-def multicall_enabled(func):
+def multicall_enabled(func: typing.Callable[..., typing.Any]) -> typing.Callable[..., typing.Any]:
     """
     Decorate the given callable to enable multicall handling.
 
     This is used by DevBuildsys methods.
 
     Args:
-        func (callable): The function to wrap.
+        func: The function to wrap.
     Returns:
-        callable: A wrapped version of func.
+        A wrapped version of func.
     """
     @wraps(func)
-    def wrapper(self, *args, **kwargs):
+    def wrapper(self, *args, **kwargs) -> typing.Any:
         """
         If multicall is enabled, store the results from func on self.
 
@@ -62,81 +69,34 @@ def multicall_enabled(func):
     return wrapper
 
 
-class Buildsystem(object):
-    """
-    The parent for our buildsystem.
-
-    Not only does this help us keep track of the functionality that we expect from our buildsystem,
-    but it also alows us to create a development subclass of this object to use during development
-    so we don't alter any production data.
-    """
-
-    def getBuild(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def getLatestBuilds(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def moveBuild(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def ssl_login(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def listBuildRPMs(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def listTags(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def listTagged(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def taskFinished(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def tagBuild(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def untagBuild(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def multiCall(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def getTag(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def addTag(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-    def deleteTag(self, *args, **kw):
-        """Raise NotImplementedError."""
-        raise NotImplementedError
-
-
-class DevBuildsys(Buildsystem):
+class DevBuildsys:
     """A dummy buildsystem instance used during development and testing."""
 
-    __untag__ = []
-    __moved__ = []
-    __added__ = []
-    __tagged__ = {}
-    __rpms__ = []
-    __tags__ = []
+    _side_tag_data = [{'id': 1234, 'name': 'f17-build-side-1234'},
+                      {'id': 7777, 'name': 'f17-build-side-7777'}]
+
+    __untag__ = []  # type: typing.List[typing.Tuple[str, str]]
+    __moved__ = []  # type: typing.List[typing.Tuple[str, str, str]]
+    __added__ = []  # type: typing.List[typing.Tuple[str, str]]
+    __tagged__ = {}  # type: typing.Mapping[str, typing.List[str]]
+    __rpms__ = []  # type: typing.List[typing.Dict[str, object]]
+    __tags__ = []  # type: typing.List[typing.Tuple[str, typing.Mapping[str, typing.Any]]]
+    __side_tags__ = _side_tag_data  # type: typing.List[typing.Dict[str, object]]
+    __removed_side_tags__ = []  # type: typing.List[typing.Dict[str, object]]
+
+    _build_data = {'build_id': 16058,
+                   'completion_time': '2007-08-24 23:26:10.890319',
+                   'completion_ts': 1187997970,
+                   'creation_event_id': 151517,
+                   'creation_time': '2007-08-24 19:38:29.422344',
+                   'extra': None,
+                   'epoch': None,
+                   'owner_id': 388,
+                   'owner_name': 'lmacken',
+                   'package_id': 8,
+                   'state': 1,
+                   'tag_id': 19,
+                   'task_id': 127621}
 
     def __init__(self):
         """Initialize the DevBuildsys."""
@@ -144,7 +104,13 @@ class DevBuildsys(Buildsystem):
         self.multicall_result = []
 
     @property
-    def multicall(self):
+    def _side_tag_ids_names(self):
+        return {id_or_name
+                for taginfo in self._side_tag_data
+                for id_or_name in (taginfo['id'], taginfo['name'])}
+
+    @property
+    def multicall(self) -> bool:
         """
         Return the value of self._multicall.
 
@@ -154,12 +120,12 @@ class DevBuildsys(Buildsystem):
         return self._multicall
 
     @multicall.setter
-    def multicall(self, value):
+    def multicall(self, value: bool):
         """
         Set the _multicall attribute to the given value.
 
         Args:
-            value (object): The value to set the _multicall attribute to.
+            value: The value to set the _multicall attribute to.
         """
         self._multicall = value
         self.multicall_result = []
@@ -172,6 +138,8 @@ class DevBuildsys(Buildsystem):
         cls.__added__ = []
         cls.__tagged__ = {}
         cls.__rpms__ = []
+        cls.__tags__ = []
+        cls.__side_tags__ = list(cls._side_tag_data)
 
     def multiCall(self):
         """Emulate Koji's multiCall."""
@@ -179,21 +147,23 @@ class DevBuildsys(Buildsystem):
         self.multicall = False
         return result
 
-    def moveBuild(self, from_tag, to_tag, build, *args, **kw):
+    def moveBuild(self, from_tag: str, to_tag: str, build: str, *args, **kw):
         """Emulate Koji's moveBuild."""
         if to_tag is None:
             raise RuntimeError('Attempt to tag {} with None.'.format(build))
         log.debug("moveBuild(%s, %s, %s)" % (from_tag, to_tag, build))
         DevBuildsys.__moved__.append((from_tag, to_tag, build))
 
-    def tagBuild(self, tag, build, *args, **kw):
+    @multicall_enabled
+    def tagBuild(self, tag: str, build: str, *args, **kw):
         """Emulate Koji's tagBuild."""
         if tag is None:
             raise RuntimeError('Attempt to tag {} with None.'.format(build))
         log.debug("tagBuild(%s, %s)" % (tag, build))
         DevBuildsys.__added__.append((tag, build))
 
-    def untagBuild(self, tag, build, *args, **kw):
+    @multicall_enabled
+    def untagBuild(self, tag: str, build: str, *args, **kw):
         """Emulate Koji's untagBuild."""
         if tag is None:
             raise RuntimeError('Attempt to untag {} with None.'.format(build))
@@ -204,21 +174,21 @@ class DevBuildsys(Buildsystem):
         """Emulate Koji's ssl_login."""
         log.debug("ssl_login(%s, %s)" % (args, kw))
 
-    def taskFinished(self, task):
+    def taskFinished(self, task: int) -> bool:
         """Emulate Koji's taskFinished."""
         return True
 
-    def getTaskInfo(self, task):
+    def getTaskInfo(self, task: int) -> typing.Mapping[str, int]:
         """Emulate Koji's getTaskInfo."""
         return {'state': koji.TASK_STATES['CLOSED']}
 
-    def getTaskRequest(self, task_id):
+    def getTaskRequest(self, task_id: int) -> typing.List[typing.Union[str, typing.Mapping]]:
         """Emulate Koji's getTaskRequest."""
         return [
-            u'git://pkgs.fedoraproject.org/rpms/bodhi?#2e994ca8b3296e62e8b0aadee1c5c0649559625a',
+            'git://pkgs.fedoraproject.org/rpms/bodhi?#2e994ca8b3296e62e8b0aadee1c5c0649559625a',
             'f17-candidate', {}]
 
-    def listPackages(self):
+    def listPackages(self) -> typing.List[typing.Mapping[str, typing.Union[int, str]]]:
         """Emulate Koji's listPackages."""
         return [
             {'package_id': 2625, 'package_name': 'nethack'},
@@ -227,25 +197,28 @@ class DevBuildsys(Buildsystem):
     @multicall_enabled
     def getBuild(self, build='TurboGears-1.0.2.2-2.fc17', other=False, testing=False):
         """Emulate Koji's getBuild."""
+        # needed to test against non-existent builds
+        if 'youdontknowme' in build:
+            return None
+
+        if 'gnome-backgrounds-3.0-1.fc17' in build:
+            return {'name': 'gnome-backgrounds',
+                    'nvr': 'gnome-backgrounds-3.0-1.fc17',
+                    'package_name': 'gnome-backgrounds',
+                    'release': '1.fc17',
+                    'tag_name': 'f17-build-side-7777',
+                    'version': '3.0'}
+
         theid = 16058
         if other and not testing:
             theid = 16059
         elif other and testing:
             theid = 16060
-        data = {'build_id': 16058,
-                'completion_time': '2007-08-24 23:26:10.890319',
-                'completion_ts': 1187997970,
-                'creation_event_id': 151517,
-                'creation_time': '2007-08-24 19:38:29.422344',
-                'extra': None,
-                'epoch': None,
-                'id': theid,
-                'owner_id': 388,
-                'owner_name': 'lmacken',
-                'package_id': 8,
-                'state': 1,
-                'tag_id': 19,
-                'task_id': 127621}
+
+        data = self._build_data.copy()
+        data['id'] = theid
+        if 'noowner' in build:
+            del data['owner_name']
 
         name, version, release = build.rsplit("-", 2)
         release_tokens = release.split(".")
@@ -327,9 +300,12 @@ class DevBuildsys(Buildsystem):
                      'tag_name': tag,
                      'version': version})
 
+        if 'testmissingnvr' in build:
+            del data['nvr']
+
         return data
 
-    def listBuildRPMs(self, id, *args, **kw):
+    def listBuildRPMs(self, id: int, *args, **kw) -> typing.List[typing.Dict[str, object]]:
         """Emulate Koji's listBuildRPMs."""
         rpms = [{'arch': 'src',
                  'build_id': 6475,
@@ -362,7 +338,7 @@ class DevBuildsys(Buildsystem):
         rpms += DevBuildsys.__rpms__
         return rpms
 
-    def listTags(self, build, *args, **kw):
+    def listTags(self, build: str, *args, **kw) -> typing.List[typing.Dict[str, object]]:
         """Emulate Koji's listTags."""
         if 'el5' in build or 'el6' in build:
             release = build.split('.')[-1].replace('el', '')
@@ -394,7 +370,7 @@ class DevBuildsys(Buildsystem):
                 {'arches': 'x86_64', 'id': 17, 'locked': True,
                  'name': 'f27M'},
             ]
-        elif 'container' in build:
+        elif build.endswith("container"):
             result = [
                 {'arches': 'x86_64', 'id': 15, 'locked': True,
                  'name': 'f28C-updates-candidate'},
@@ -413,7 +389,7 @@ class DevBuildsys(Buildsystem):
                  'name': 'f28F'},
             ]
         else:
-            release = build.split('.')[-1].replace('fc', 'f')
+            release = build.split('.')[-1].replace('fc', 'f').replace('~bootstrap', '')
             result = [
                 {'arches': 'i386 x86_64 ppc ppc64', 'id': 10, 'locked': True,
                  'name': '%s-updates-candidate' % release, 'perm': None, 'perm_id': None},
@@ -427,12 +403,23 @@ class DevBuildsys(Buildsystem):
         return result
 
     @multicall_enabled
-    def listTagged(self, tag, *args, **kw):
-        """List updates tagged with teh given tag."""
+    def listTagged(self, tag: str, *args, **kw) -> typing.List[typing.Any]:
+        """List updates tagged with the given tag."""
+        latest = kw.get('latest', False)
+        if tag in self._side_tag_ids_names:
+            return [self.getBuild(build="gnome-backgrounds-3.0-1.fc17")]
         builds = []
-        for build in [self.getBuild(),
+
+        all_builds = [self.getBuild(),
                       self.getBuild(other=True),
-                      self.getBuild(other=True, testing=True)]:
+                      self.getBuild(other=True, testing=True)]
+
+        if latest:
+            # Delete all older builds which aren't the latest for their tag.
+            # Knowing which these are is simpler than trying to rpmvercmp.
+            del all_builds[0]
+
+        for build in all_builds:
             if build['nvr'] in self.__untag__:
                 log.debug('Pruning koji build %s' % build['nvr'])
                 continue
@@ -442,10 +429,9 @@ class DevBuildsys(Buildsystem):
             for tag_ in DevBuildsys.__tagged__[build]:
                 if tag_ == tag:
                     builds.append(self.getBuild(build))
-        log.debug(builds)
         return builds
 
-    def getLatestBuilds(self, *args, **kw):
+    def getLatestBuilds(self, *args, **kw) -> typing.List[typing.Any]:
         """
         Return a list of the output from self.getBuild().
 
@@ -454,12 +440,13 @@ class DevBuildsys(Buildsystem):
         """
         return [self.getBuild()]
 
+    @multicall_enabled
     def getTag(self, taginfo, **kw):
         """
         Retrieve the given tag from koji.
 
         Args:
-            taginfo (int or basestring): The tag you want info about.
+            taginfo (int or str): The tag you want info about.
             strict (bool): If True, raise an Exception if epel tags are queried. Defaults to False.
         Returns:
             dict or None: A dictionary of tag information, or None if epel is requested and strict
@@ -467,12 +454,6 @@ class DevBuildsys(Buildsystem):
         Raises:
             koji.GenericError: If strict is True and epel is requested.
         """
-        for nr in self.__tags__:
-            if taginfo == self.__tags__[nr][0]:
-                toreturn = self.__tags__[nr][1]
-                toreturn['id'] = nr
-                return toreturn
-
         if isinstance(taginfo, int):
             taginfo = "f%d" % taginfo
 
@@ -483,11 +464,56 @@ class DevBuildsys(Buildsystem):
             else:
                 return None
 
+        # These tags needs to be created
+        if taginfo in ["f32-build-side-1234-signing-pending",
+                       "f32-build-side-1234-testing-pending"]:
+            return None
+
+        # emulate a side-tag response
+        if taginfo in self._side_tag_ids_names:
+            for sidetag in self.__side_tags__:
+                if taginfo in (sidetag['id'], sidetag['name']):
+                    return {'maven_support': False, 'locked': False, 'name': sidetag['name'],
+                            'extra': {'sidetag_user': 'dudemcpants', 'sidetag': True},
+                            'perm': None, 'perm_id': None, 'arches': None,
+                            'maven_include_all': False, 'id': sidetag['id']}
+
+            if kw.get('strict'):
+                raise koji.GenericError("Invalid tagInfo: '%s'" % taginfo)
+            else:
+                return None
+
         return {'maven_support': False, 'locked': False, 'name': taginfo,
-                'perm': None, 'id': 246, 'arches': None,
+                'extra': {}, 'perm': None, 'id': 246, 'arches': None,
                 'maven_include_all': False, 'perm_id': None}
 
-    def addTag(self, tag, **opts):
+    def getFullInheritance(self, taginfo, **kw):
+        """
+        Return a tag inheritance.
+
+        Args:
+            taginfo (int or str): The tag. does not impact the output
+        Returns:
+            list: A list of dicts of tag information
+        """
+        return [{'intransitive': False, 'name': 'f17-build', 'pkg_filter': '', 'priority': 0,
+                 'parent_id': 6448, 'maxdepth': None, 'noconfig': False, 'child_id': 7715,
+                 'nextdepth': None, 'filter': [], 'currdepth': 1},
+                {'intransitive': False, 'name': 'f17-override', 'pkg_filter': '', 'priority': 0,
+                 'parent_id': 6447, 'maxdepth': None, 'noconfig': False, 'child_id': 6448,
+                 'nextdepth': None, 'filter': [], 'currdepth': 2},
+                {'intransitive': False, 'name': 'f17-updates', 'pkg_filter': '', 'priority': 0,
+                 'parent_id': 6441, 'maxdepth': None, 'noconfig': False, 'child_id': 6447,
+                 'nextdepth': None, 'filter': [], 'currdepth': 3},
+                {'intransitive': False, 'name': 'f17', 'pkg_filter': '', 'priority': 0,
+                 'parent_id': 6438, 'maxdepth': None, 'noconfig': False, 'child_id': 6441,
+                 'nextdepth': None, 'filter': [], 'currdepth': 4}]
+
+    def listSideTags(self, **kw):
+        """Return a list of side-tags."""
+        return self.__side_tags__
+
+    def createTag(self, tag: str, **opts):
         """Emulate tag adding."""
         if 'parent' not in opts:
             raise ValueError('No parent in tag options')
@@ -503,19 +529,30 @@ class DevBuildsys(Buildsystem):
         opts['perm_id'] = 1
         self.__tags__.append((tag, opts))
 
-    def deleteTag(self, tagid):
-        """Emulate tag deletion."""
-        del self.__tags__[tagid]
+    def editTag2(self, *args, **kw):
+        """Edit a tag."""
+        pass
 
-    def getRPMHeaders(self, rpmID, headers):
+    def deleteTag(self, tagid: typing.Union[str, int]):
+        """Emulate tag deletion."""
+        if isinstance(tagid, str):
+            for tid, tinfo in self.__tags__:
+                if tagid == tid:
+                    self.__tags__.remove((tid, tinfo))
+                    return
+        else:
+            del self.__tags__[tagid]
+
+    def getRPMHeaders(self, rpmID: str,
+                      headers: typing.Any) -> typing.Union[typing.Mapping[str, str], None]:
         """
         Return headers for the given RPM.
 
         Args:
-            rpmID (basestring): The RPM you want headers for.
-            headers (object): Unused.
+            rpmID: The RPM you want headers for.
+            headers: Unused.
         Returns:
-            dict: A dictionary of RPM headers.
+            A dictionary of RPM headers, or None if the rpmID is not found.
         """
         if rpmID == 'raise-exception.src':
             raise Exception
@@ -573,15 +610,16 @@ class DevBuildsys(Buildsystem):
             return headers
 
 
-def koji_login(config, authenticate):
+@backoff.on_exception(backoff.expo, koji.AuthError, max_time=600)
+def koji_login(config: 'BodhiConfig', authenticate: bool) -> koji.ClientSession:
     """
     Login to Koji and return the session.
 
     Args:
-        config (bodhi.server.config.BodhiConfig): Bodhi's configuration dictionary.
-        authenticate (bool): If True, establish an authenticated client session.
+        config: Bodhi's configuration dictionary.
+        authenticate: If True, establish an authenticated client session.
     Returns:
-        koji.ClientSession: An authenticated Koji ClientSession that is ready to use.
+        An authenticated Koji ClientSession that is ready to use.
     """
     koji_options = {
         'krb_rdns': False,
@@ -593,19 +631,19 @@ def koji_login(config, authenticate):
     }
 
     koji_client = koji.ClientSession(_koji_hub, koji_options)
-    if authenticate and not koji_client.krb_login(**get_krb_conf(config)):
-        log.error('Koji krb_login failed')
+    if authenticate and not koji_client.gssapi_login(**get_krb_conf(config)):
+        log.error('Koji gssapi_login failed')
     return koji_client
 
 
-def get_krb_conf(config):
+def get_krb_conf(config: 'BodhiConfig') -> typing.Mapping[str, str]:
     """
-    Return arguments for krb_login.
+    Return arguments for gssapi_login.
 
     Args:
-        config (bodhi.server.config.BodhiConfig): Bodhi's configuration dictionary.
+        config: Bodhi's configuration dictionary.
     Returns:
-        dict: A dictionary containing three keys:
+        A dictionary containing three keys:
             principal: The kerberos principal to use.
             keytab: The kerberos keytab to use.
             ccache: The kerberos ccache to use.
@@ -619,20 +657,21 @@ def get_krb_conf(config):
     if keytab:
         args['keytab'] = keytab
     if ccache:
+        ccache = ccache.replace('%{uid}', str(os.geteuid()))
         args['ccache'] = ccache
     return args
 
 
-def get_session():
+def get_session() -> typing.Union[koji.ClientSession, DevBuildsys]:
     """
     Get a new buildsystem instance.
 
     Returns:
-        koji.ClientSession or DevBuildsys: A buildsystem client instance.
+        A buildsystem client instance.
     Raises:
         RuntimeError: If the build system has not been initialized. See setup_buildsystem().
     """
-    global _buildsystem, _buildsystem_login_lock
+    global _buildsystem
     if _buildsystem is None:
         raise RuntimeError('Buildsys needs to be setup')
     with _buildsystem_login_lock:
@@ -646,21 +685,20 @@ def teardown_buildsystem():
     DevBuildsys.clear()
 
 
-def setup_buildsystem(settings, authenticate=True):
+def setup_buildsystem(settings: 'BodhiConfig', authenticate: bool = True):
     """
     Initialize the buildsystem client.
 
     Args:
-        settings (bodhi.server.config.BodhiConfig): Bodhi's config.
-        authenticate (bool): If True, establish an authenticated Koji session. Defaults to True.
+        settings: Bodhi's config.
+        authenticate: If True, establish an authenticated Koji session. Defaults to True.
     Raises:
         ValueError: If the buildsystem is configured to an invalid value.
     """
-    global _buildsystem, _koji_hub, _buildsystem_login_lock
+    global _buildsystem, _koji_hub
     if _buildsystem:
         return
 
-    _buildsystem_login_lock = Lock()
     _koji_hub = settings.get('koji_hub')
     buildsys = settings.get('buildsystem')
 
@@ -679,17 +717,20 @@ def setup_buildsystem(settings, authenticate=True):
         raise ValueError('Buildsys %s not known' % buildsys)
 
 
-def wait_for_tasks(tasks, session=None, sleep=300):
+def wait_for_tasks(
+        tasks: typing.List[typing.Any],
+        session: typing.Union[koji.ClientSession, None] = None,
+        sleep: int = 300) -> typing.List[typing.Any]:
     """
     Wait for a list of koji tasks to complete.
 
     Args:
-        tasks (list): The return value of Koji's multiCall().
-        session (koji.ClientSession or None): A Koji client session to use. If not provided, the
+        tasks: The return value of Koji's multiCall().
+        session: A Koji client session to use. If not provided, the
             function will acquire its own session.
-        sleep (int): How long to sleep between polls on Koji when waiting for tasks to complete.
+        sleep: How long to sleep between polls on Koji when waiting for tasks to complete.
     Returns:
-        list: A list of failed tasks. An empty list indicates that all tasks completed successfully.
+        A list of failed tasks. An empty list indicates that all tasks completed successfully.
     """
     log.debug("Waiting for %d tasks to complete: %s" % (len(tasks), tasks))
     failed_tasks = []
