@@ -1,5 +1,4 @@
-# -*- coding: utf-8 -*-
-# Copyright © 2013-2017 Red Hat, Inc. and others.
+# Copyright © 2013-2019 Red Hat, Inc. and others.
 #
 # This file is part of Bodhi.
 #
@@ -18,34 +17,34 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 """Bodhi's configuration and configuration loading and validation mechanisms."""
 from datetime import datetime
-import os
 import logging
-import binascii
+import os
+import typing
 
 from pyramid import settings
 from pyramid.paster import get_appsettings
-import cryptography.fernet
-import six
 
 
 log = logging.getLogger('bodhi')
 
 
-def get_configfile():
+def get_configfile() -> typing.Optional[str]:
     """
     Return a path to a config file, if found.
 
-    Return the path to a config file, with a heirarchy of preferential paths. It searches first
-    for development.ini if found. if not found, it will return /etc/bodhi/production.ini if it
+    Return the path to a config file, with a hierarchy of preferential paths. It searches first
+    for development.ini if found. If not found, it will return /etc/bodhi/production.ini if it
     exists. Otherwise, it returns None.
 
     Returns:
-        basestring or None: The path of a config file, or None if no config file is found.
+        The path of a config file, or None if no config file is found.
     """
     configfile = None
     setupdir = os.path.join(os.path.dirname(os.path.dirname(__file__)), '..')
-    for cfg in (os.path.join(setupdir, 'development.ini'),
-                '/etc/bodhi/production.ini'):
+    possible_configs = [os.path.join(setupdir, 'development.ini'), '/etc/bodhi/production.ini']
+    if "BODHI_CONFIG" in os.environ:
+        possible_configs.insert(0, os.environ["BODHI_CONFIG"])
+    for cfg in possible_configs:
         if os.path.exists(cfg):
             configfile = cfg
             break
@@ -54,7 +53,40 @@ def get_configfile():
     return configfile
 
 
-def _generate_list_validator(splitter=' ', validator=six.text_type):
+def _generate_dict_validator(value: str) -> dict:
+    """Return a dict version of value.
+
+    This function accept a string with comma separated tuples in the form `key:value`
+    and ensure it can be correctly imported as a dict. It will return a dict.
+    If it cannot do that, it will raise ValueError.
+
+    Args:
+        value: The value to be validated as a dict.
+    Returns:
+        The dict interpretation of value.
+    Raises:
+        ValueError: If value cannot be interpreted as a dict.
+    """
+    if isinstance(value, str):
+        values_list = [idx.strip() for idx in value.split(',') if idx.strip()]
+        dictvalue = dict()
+        for v in values_list:
+            try:
+                k, v = v.split(':')
+                dictvalue[k.strip()] = v.strip()
+            except Exception:
+                raise ValueError(f'"{value}" cannot be interpreted as a dict.')
+        value = dictvalue
+
+    if not isinstance(value, dict):
+        raise ValueError(f'"{value}" cannot be interpreted as a dict.')
+
+    return value
+
+
+def _generate_list_validator(
+        splitter: str = ' ', validator: typing.Callable[[typing.Any], typing.Any] = str) \
+        -> typing.Callable[[typing.Union[str, typing.List]], typing.Any]:
     """Return a function that takes a value and interprets it to be a list with the given splitter.
 
     This function generates a function that can take a string and interpret it as a list by
@@ -62,12 +94,12 @@ def _generate_list_validator(splitter=' ', validator=six.text_type):
     with the given validator.
 
     Args:
-        splitter (basestring): A string to use to split the input into a list.
-        validator (function):  A function to apply to each element of the list to validate it.
+        splitter: A string to use to split the input into a list.
+        validator:  A function to apply to each element of the list to validate it.
     Returns:
-        function: A validator function that accepts an argument to be validated.
+        A validator function that accepts an argument to be validated.
     """
-    def _validate_list(value):
+    def _validate_list(value: typing.Union[str, typing.List]) -> typing.List:
         """Validate that the value is a list or can be split into a list, and validate its elements.
 
         This function will validate that the given value is a list, or it will use the splitter to
@@ -75,17 +107,17 @@ def _generate_list_validator(splitter=' ', validator=six.text_type):
         list.
 
         Args:
-            value (basestring or list): The list to be validated.
+            value: The list to be validated.
         Returns:
-            unicode: The interpreted list.
+            The interpreted list.
         Raises:
             ValueError: If validator fails on any of the list's elements.
         """
-        if isinstance(value, six.string_types):
+        if isinstance(value, str):
             value = [idx.strip() for idx in value.split(splitter) if idx.strip()]
 
         if not isinstance(value, list):
-            raise ValueError('"{}" cannot be intepreted as a list.'.format(value))
+            raise ValueError('"{}" cannot be interpreted as a list.'.format(value))
 
         # Run the validator on each element of the list.
         value = [validator(v) for v in value]
@@ -95,20 +127,20 @@ def _generate_list_validator(splitter=' ', validator=six.text_type):
     return _validate_list
 
 
-def _validate_bool(value):
+def _validate_bool(value: typing.Union[str, bool]) -> bool:
     """Return a bool version of value.
 
     This function will ensure that value is a bool, or that it is a string that can be interpreted
     as a bool. It will return a bool. If it cannot do that, it will raise ValueError.
 
     Args:
-        value (basestring or bool): The value to be validated as a bool.
+        value: The value to be validated as a bool.
     Returns:
-        bool: The boolean interpretation of value.
+        The boolean interpretation of value.
     Raises:
         ValueError: If value cannot be interpreted as a boolean.
     """
-    if isinstance(value, six.string_types):
+    if isinstance(value, str):
         # Recent versions of Pyramid define a settings.falsey, but version 1.5.6 does not so its
         # values have been backported here for the False case. Pyramid defines an asbool(), but it
         # will not raise any error for strings that aren't in the truthy or falsey lists, and we
@@ -126,75 +158,19 @@ def _validate_bool(value):
     return value
 
 
-def _validate_color(value):
-    """Ensure that value is a valid expression of a color, in the form #dddddd.
-
-    Return the value if it is a valid color expression, or raise ValueError.
-
-    Args:
-        value (basestring): The color to be validated.
-    Returns:
-        unicode: The color.
-    Raises:
-        ValueError: If value is not in the form #dddddd.
-    """
-    e = ValueError('"{}" is not a valid color expression.'.format(value))
-
-    if not isinstance(value, six.string_types):
-        raise e
-    if not len(value) == 7:
-        raise e
-    if value[0] != '#':
-        raise e
-    try:
-        int(value[-6:], 16)
-    except ValueError:
-        raise e
-
-    return six.text_type(value)
-
-
-def _validate_fernet_key(value):
-    """Ensure the value is not CHANGEME, that it is a Fernet key, and convert it to a str.
-
-    This function is used to ensure that secret values in the config have been set by the user to
-    something other than the default of CHANGEME and that the value can be used as a Fernet key. It
-    is converted to str before returning.
-
-    Args:
-        value (basestring): The value to be validated.
-    Returns:
-        str: The value.
-    Raises:
-        ValueError: If value is "CHANGEME" or if it cannot be used as a Fernet key.
-    """
-    _validate_secret(value)
-
-    if isinstance(value, six.text_type):
-        value = value.encode('utf-8')
-
-    try:
-        engine = cryptography.fernet.Fernet(value)
-        # This will raise a ValueError if value is not suitable as a Fernet key.
-        engine.encrypt(b'a secret test string')
-    except (TypeError, binascii.Error):
-        raise ValueError('Fernet key must be 32 url-safe base64-encoded bytes.')
-
-    return value
-
-
-def _validate_none_or(validator):
+def _validate_none_or(validator: typing.Callable[[typing.Any], typing.Any]) \
+        -> typing.Callable[[typing.Any], typing.Any]:
     """Return a function that will ensure a value is None or passes validator.
 
     This function returns a function that will take a single argument, value, and will ensure
     that value is None or that it passes the given validator.
 
     Args:
-        validator (function): A function to apply when value is not None.
+        validator: A function to apply when value is not None.
     Returns:
-        function: A validator function that accepts an argument to be validated.
+        A validator function that accepts an argument to be validated.
     """
-    def _validate(value):
+    def _validate(value: typing.Any) -> typing.Any:
         if value is None:
             return value
 
@@ -203,71 +179,71 @@ def _validate_none_or(validator):
     return _validate
 
 
-def validate_path(value):
+def validate_path(value: str) -> str:
     """Ensure that value is an existing path on the local filesystem and return it.
 
     Use os.path.exists to ensure that value is an existing path. Return the value if it is, else
     raise ValueError.
 
     Args:
-        value (basestring): The path to be validated.
+        value: The path to be validated.
     Returns:
-        unicode: The path.
+        The path.
     Raises:
         ValueError: If os.path.exists returns False.
     """
-    if not os.path.exists(value):
-        raise ValueError('"{}" does not exist.'.format(value))
+    if value is None or not os.path.exists(value):
+        raise ValueError(f'{value!r} does not exist.')
 
-    return six.text_type(value)
+    return str(value)
 
 
-def _validate_rstripped_str(value):
+def _validate_rstripped_str(value: str) -> str:
     """
     Ensure that value is a str that is rstripped of the / character.
 
     Args:
-        value (six.text_type): The value to be validated and rstripped.
+        value: The value to be validated and rstripped.
     Returns:
-        six.text_type: The rstripped value.
+        The rstripped value.
     """
-    value = six.text_type(value)
+    value = str(value)
     return value.rstrip('/')
 
 
-def _validate_secret(value):
-    """Ensure that the value is not CHANGEME and convert it to unicode.
+def _validate_secret(value: str) -> str:
+    """Ensure that the value is not CHANGEME and convert it to a string.
 
     This function is used to ensure that secret values in the config have been set by the user to
     something other than the default of CHANGEME.
 
     Args:
-        value (basestring): The value to be validated.
+        value: The value to be validated.
     Returns:
-        unicode: The value.
+        The value.
     Raises:
         ValueError: If value is "CHANGEME".
     """
     if value == 'CHANGEME':
         raise ValueError('This setting must be changed from its default value.')
 
-    return six.text_type(value)
+    return str(value)
 
 
-def _validate_tls_url(value):
+def _validate_tls_url(value: str) -> str:
     """Ensure that the value is a string that starts with https://.
 
     Args:
-        value (basestring): The value to be validated.
+        value: The value to be validated.
     Returns:
-        unicode: The value.
+        The value.
     Raises:
         ValueError: If value is not a string starting with https://.
     """
-    if not isinstance(value, six.string_types) or not value.startswith('https://'):
+    if not isinstance(value, str) or not value.startswith('https://'):
         raise ValueError('This setting must be a URL starting with https://.')
 
-    return six.text_type(value)
+    return str(value)
 
 
 class BodhiConfig(dict):
@@ -283,14 +259,16 @@ class BodhiConfig(dict):
     _defaults = {
         'acl_system': {
             'value': 'dummy',
-            'validator': six.text_type},
+            'validator': str},
         'acl_dummy_committer': {
             'value': None,
-            'validator': _validate_none_or(six.text_type)},
+            'validator': _validate_none_or(str)},
         'admin_groups': {
+            # Defined in and tied to the Fedora Account System (limited to 16 characters)
             'value': ['proventesters', 'security_respons', 'bodhiadmin', 'sysadmin-main'],
             'validator': _generate_list_validator()},
         'admin_packager_groups': {
+            # Defined in and tied to the Fedora Account System (limited to 16 characters)
             'value': ['provenpackager', 'releng', 'security_respons'],
             'validator': _generate_list_validator()},
         'authtkt.secret': {
@@ -302,87 +280,77 @@ class BodhiConfig(dict):
         'authtkt.timeout': {
             'value': 86400,
             'validator': int},
-        'badge_ids': {
-            'value': [],
-            'validator': _generate_list_validator('|')},
+        'automatic_updates_blacklist': {
+            # List of users to not create automatic updates from
+            'value': ['releng'],
+            'validator': _generate_list_validator()},
         'base_address': {
             'value': 'https://admin.fedoraproject.org/updates/',
-            'validator': six.text_type},
+            'validator': str},
         'bodhi_email': {
             'value': 'updates@fedoraproject.org',
-            'validator': six.text_type},
+            'validator': str},
         'bodhi_password': {
             'value': None,
-            'validator': _validate_none_or(six.text_type)},
+            'validator': _validate_none_or(str)},
         'buglink': {
             'value': 'https://bugzilla.redhat.com/show_bug.cgi?id=%s',
-            'validator': six.text_type},
+            'validator': str},
         'bugtracker': {
             'value': None,
-            'validator': _validate_none_or(six.text_type)},
+            'validator': _validate_none_or(str)},
         'bugzilla_api_key': {
             'value': None,
-            'validator': _validate_none_or(six.text_type)},
+            'validator': _validate_none_or(str)},
         'buildroot_limit': {
             'value': 31,
             'validator': int},
         'buildsystem': {
             'value': 'dev',
-            'validator': six.text_type},
-        'bz_products': {
+            'validator': str},
+        'bz_exclude_rels': {
             'value': [],
             'validator': _generate_list_validator(',')},
+        'bz_products': {
+            'value': ['Fedora', 'Fedora EPEL', 'Fedora Modules'],
+            'validator': _generate_list_validator(',')},
+        'bz_regex': {
+            'value': (r'(?:fix(?:es)?|close(?:s)?|resolve(?:s)?)(?:\:)?\s'
+                      r'(?:fedora|epel|rh(?:bz)?)#(\d{5,})'),
+            'validator': str},
         'bz_server': {
             'value': 'https://bugzilla.redhat.com/xmlrpc.cgi',
-            'validator': six.text_type},
+            'validator': str},
+        'bz_server_rest': {
+            'value': 'https://bugzilla.redhat.com/rest/',
+            'validator': str},
         'cache_dir': {
             'value': None,
             'validator': _validate_none_or(validate_path)},
-        'captcha.background_color': {
-            'value': '#ffffff',
-            'validator': _validate_color},
-        'captcha.font_color': {
-            'value': '#000000',
-            'validator': _validate_color},
-        'captcha.font_path': {
-            'value': '/usr/share/fonts/liberation/LiberationMono-Regular.ttf',
-            'validator': validate_path},
-        'captcha.font_size': {
-            'value': 36,
-            'validator': int},
-        'captcha.image_height': {
-            'value': 80,
-            'validator': int},
-        'captcha.image_width': {
-            'value': 300,
-            'validator': int},
-        'captcha.padding': {
-            'value': 5,
-            'validator': int},
-        'captcha.secret': {
-            'value': None,
-            'validator': _validate_none_or(_validate_fernet_key)},
-        'captcha.ttl': {
-            'value': 300,
+        'celery_config': {
+            'value': '/etc/bodhi/celeryconfig.py',
+            'validator': str},
+        'check_signed_builds_delay': {
+            'value': 2,
             'validator': int},
         'clean_old_composes': {
             'value': True,
             'validator': _validate_bool},
         'container.destination_registry': {
             'value': 'registry.fedoraproject.org',
-            'validator': six.text_type},
+            'validator': str},
         'container.source_registry': {
             'value': 'candidate-registry.fedoraproject.org',
-            'validator': six.text_type},
+            'validator': str},
         'cors_connect_src': {
             'value': 'https://*.fedoraproject.org/ wss://hub.fedoraproject.org:9939/',
-            'validator': six.text_type},
+            'validator': str},
         'cors_origins_ro': {
             'value': '*',
-            'validator': six.text_type},
+            'validator': str},
         'cors_origins_rw': {
             'value': 'https://bodhi.fedoraproject.org',
-            'validator': six.text_type},
+            'validator': str},
         'critpath_pkgs': {
             'value': [],
             'validator': _generate_list_validator()},
@@ -397,61 +365,53 @@ class BodhiConfig(dict):
             'validator': int},
         'critpath.type': {
             'value': None,
-            'validator': _validate_none_or(six.text_type)},
-        'datagrepper_url': {
-            'value': 'https://apps.fedoraproject.org/datagrepper',
-            'validator': six.text_type},
+            'validator': _validate_none_or(str)},
         'default_email_domain': {
             'value': 'fedoraproject.org',
-            'validator': six.text_type},
+            'validator': str},
         'disable_automatic_push_to_stable': {
             'value': (
                 'Bodhi is disabling automatic push to stable due to negative karma. The '
                 'maintainer may push manually if they determine that the issue is not severe.'),
-            'validator': six.text_type},
+            'validator': str},
         'dogpile.cache.arguments.filename': {
             'value': '/var/cache/bodhi-dogpile-cache.dbm',
-            'validator': six.text_type},
+            'validator': str},
         'dogpile.cache.backend': {
             'value': 'dogpile.cache.dbm',
-            'validator': six.text_type},
+            'validator': str},
         'dogpile.cache.expiration_time': {
             'value': 100,
             'validator': int},
         'exclude_mail': {
             'value': ['autoqa', 'taskotron'],
             'validator': _generate_list_validator()},
-        'fedmenu.data_url': {
-            'value': 'https://apps.fedoraproject.org/js/data.js',
-            'validator': six.text_type},
-        'fedmenu.url': {
-            'value': 'https://apps.fedoraproject.org/fedmenu',
-            'validator': six.text_type},
-        'fedmsg_enabled': {
-            'value': False,
-            'validator': _validate_bool},
         'file_url': {
             'value': 'https://download.fedoraproject.org/pub/fedora/linux/updates',
-            'validator': six.text_type},
+            'validator': str},
         'fmn_url': {
             'value': 'https://apps.fedoraproject.org/notifications/',
-            'validator': six.text_type},
+            'validator': str},
         'important_groups': {
-            'value': ['proventesters', 'provenpackager,' 'releng', 'security_respons', 'packager',
+            # Defined in and tied to the Fedora Account System (limited to 16 characters)
+            'value': ['proventesters', 'provenpackager', 'releng', 'security_respons', 'packager',
                       'bodhiadmin'],
             'validator': _generate_list_validator()},
         'initial_bug_msg': {
             'value': '%s has been submitted as an update to %s. %s',
-            'validator': six.text_type},
+            'validator': str},
         'greenwave_api_url': {
             'value': 'https://greenwave-web-greenwave.app.os.fedoraproject.org/api/v1.0',
             'validator': _validate_rstripped_str},
+        'greenwave_batch_size': {
+            'value': 8,
+            'validator': int},
         'waiverdb_api_url': {
             'value': 'https://waiverdb-web-waiverdb.app.os.fedoraproject.org/api/v1.0',
             'validator': _validate_rstripped_str},
         'waiverdb.access_token': {
             'value': None,
-            'validator': _validate_none_or(six.text_type)},
+            'validator': _validate_none_or(str)},
         'koji_web_url': {
             'value': 'https://koji.fedoraproject.org/koji/',
             'validator': _validate_tls_url},
@@ -469,139 +429,142 @@ class BodhiConfig(dict):
             'validator': _validate_none_or(str)},
         'legal_link': {
             'value': '',
-            'validator': six.text_type},
+            'validator': str},
         'libravatar_dns': {
             'value': False,
             'validator': _validate_bool},
         'libravatar_enabled': {
             'value': True,
             'validator': _validate_bool},
+        'libravatar_prefer_tls': {
+            'value': True,
+            'validator': bool},
         'mail.templates_basepath': {
-            'value': 'bodhi:server/email/templates/',
-            'validator': six.text_type},
+            'value': 'bodhi.server:email/templates/',
+            'validator': str},
         'mako.directories': {
-            'value': 'bodhi:server/templates',
-            'validator': six.text_type},
+            'value': 'bodhi.server:templates',
+            'validator': str},
         'mandatory_packager_groups': {
             'value': ['packager'],
             'validator': _generate_list_validator()},
-        'mash_dir': {
+        'compose_dir': {
             'value': None,
-            'validator': _validate_none_or(six.text_type)},
-        'mash_stage_dir': {
+            'validator': _validate_none_or(str)},
+        'compose_stage_dir': {
             'value': None,
-            'validator': _validate_none_or(six.text_type)},
-        'max_concurrent_mashes': {
+            'validator': _validate_none_or(str)},
+        'max_concurrent_composes': {
             'value': 2,
             'validator': int},
-        'max_update_length_for_ui': {
-            'value': 30,
-            'validator': int},
         'message_id_email_domain': {
             'value': 'admin.fedoraproject.org',
-            'validator': six.text_type},
+            'validator': str},
         'not_yet_tested_epel_msg': {
             'value': (
                 'This update has not yet met the minimum testing requirements defined in the '
                 '<a href="https://fedoraproject.org/wiki/EPEL_Updates_Policy">EPEL Update Policy'
                 '</a>'),
-            'validator': six.text_type},
+            'validator': str},
         'not_yet_tested_msg': {
             'value': (
                 'This update has not yet met the minimum testing requirements defined in the '
                 '<a href="https://fedoraproject.org/wiki/Package_update_acceptance_criteria">'
                 'Package Update Acceptance Criteria</a>'),
-            'validator': six.text_type},
+            'validator': str},
         'openid.provider': {
             'value': 'https://id.fedoraproject.org/openid/',
-            'validator': six.text_type},
+            'validator': str},
         'openid.sreg_required': {
-            'value': 'email',
-            'validator': six.text_type},
+            'value': 'email nickname',
+            'validator': str},
         'openid.success_callback': {
-            'value': 'bodhi.server.security:remember_me',
-            'validator': six.text_type},
+            'value': 'bodhi.server.auth.utils:remember_me',
+            'validator': str},
         'openid.url': {
             'value': 'https://id.fedoraproject.org/',
-            'validator': six.text_type},
+            'validator': str},
         'openid_template': {
             'value': '{username}.id.fedoraproject.org',
-            'validator': six.text_type},
-        'pagure_flatpak_namespace': {
-            'value': 'modules',
-            'validator': six.text_type},
+            'validator': str},
+        'oidc.fedora.client_id': {
+            'value': '',
+            'validator': str},
+        'oidc.fedora.client_secret': {
+            'value': '',
+            'validator': str},
+        'oidc.fedora.server_metadata_url': {
+            'value': 'https://id.fedoraproject.org/openidc/.well-known/openid-configuration',
+            'validator': str},
+        'pagure_namespaces': {
+            'value': ('rpm:rpms, module:modules, container:container, flatpak:flatpaks'),
+            'validator': _generate_dict_validator},
+        'pagure_flatpak_main_branch': {
+            'value': 'stable',
+            'validator': str},
+        'pagure_module_main_branch': {
+            'value': 'master',
+            'validator': str},
         'pagure_url': {
             'value': 'https://src.fedoraproject.org/pagure/',
             'validator': _validate_tls_url},
         'pdc_url': {
             'value': 'https://pdc.fedoraproject.org/',
             'validator': _validate_tls_url},
-        'pkgdb_url': {
-            'value': 'https://admin.fedoraproject.org/pkgdb',
-            'validator': six.text_type},
-        'prefer_ssl': {
-            'value': None,
-            'validator': _validate_none_or(bool)},
         'privacy_link': {
             'value': '',
-            'validator': six.text_type},
+            'validator': str},
         'pungi.basepath': {
             'value': '/etc/bodhi',
-            'validator': six.text_type},
+            'validator': str},
         'pungi.cmd': {
             'value': '/usr/bin/pungi-koji',
-            'validator': six.text_type},
+            'validator': str},
         'pungi.conf.module': {
             'value': 'pungi.module.conf',
-            'validator': six.text_type},
+            'validator': str},
         'pungi.conf.rpm': {
             'value': 'pungi.rpm.conf',
-            'validator': six.text_type},
+            'validator': str},
         'pungi.extracmdline': {
             'value': [],
             'validator': _generate_list_validator()},
         'pungi.labeltype': {
             'value': 'Update',
-            'validator': six.text_type},
+            'validator': str},
         'query_wiki_test_cases': {
             'value': False,
             'validator': _validate_bool},
         'release_team_address': {
             'value': 'bodhiadmin-members@fedoraproject.org',
-            'validator': six.text_type},
+            'validator': str},
         'resultsdb_api_url': {
             'value': 'https://taskotron.fedoraproject.org/resultsdb_api/',
-            'validator': six.text_type},
+            'validator': str},
         'session.secret': {
             'value': 'CHANGEME',
             'validator': _validate_secret},
         'site_requirements': {
-            'value': 'dist.rpmdeplint dist.upgradepath',
-            'validator': six.text_type},
+            'value': 'dist.rpmdeplint',
+            'validator': str},
         'skopeo.cmd': {
             'value': '/usr/bin/skopeo',
-            'validator': six.text_type,
+            'validator': str,
         },
         'skopeo.extra_copy_flags': {
             'value': '',
-            'validator': six.text_type,
+            'validator': str,
         },
         'smtp_server': {
             'value': None,
-            'validator': _validate_none_or(six.text_type)},
+            'validator': _validate_none_or(str)},
         'sqlalchemy.url': {
-            'value': 'sqlite:////var/cache/bodhi.db',
-            'validator': six.text_type},
+            'value': 'postgresql://localhost/bodhi',
+            'validator': str},
         'stable_bug_msg': {
-            'value': ('%s has been pushed to the %s repository. If problems still persist, please '
-                      'make note of it in this bug report.'),
-            'validator': six.text_type},
-        'stable_from_batched_msg': {
-            'value': ('This update has been dequeued from batched and is now entering stable.'),
-            'validator': six.text_type},
-        'stacks_enabled': {
-            'value': False,
-            'validator': _validate_bool},
+            'value': ('{update_alias} has been pushed to the {repo} repository.\n'
+                      'If problem still persists, please make note of it in this bug report.'),
+            'validator': str},
         'stats_blacklist': {
             'value': ['bodhi', 'anonymous', 'autoqa', 'taskotron'],
             'validator': _generate_list_validator()},
@@ -610,26 +573,26 @@ class BodhiConfig(dict):
             'validator': _generate_list_validator()},
         'test_case_base_url': {
             'value': 'https://fedoraproject.org/wiki/',
-            'validator': six.text_type},
-        'testing_approval_msg_based_on_karma': {
-            'value': ('This update has reached the stable karma threshold and can be pushed to '
-                      'stable now if the maintainer wishes.'),
-            'validator': six.text_type
-        },
+            'validator': str},
         'testing_approval_msg': {
-            'value': ('This update has reached %d days in testing and can be pushed to stable now '
-                      'if the maintainer wishes'),
-            'validator': six.text_type},
+            'value': ('This update can be pushed to stable now if the maintainer wishes'),
+            'validator': str},
         'testing_bug_epel_msg': {
             'value': (
-                '\nSee https://fedoraproject.org/wiki/QA:Updates_Testing for\ninstructions on how '
-                'to install test updates.\nYou can provide feedback for this update here: %s'),
-            'validator': six.text_type},
+                '{update_alias} has been pushed to the {repo} repository.\n'
+                '{install_instructions}\n'
+                'You can provide feedback for this update here: {update_url}\n\n'
+                'See also https://fedoraproject.org/wiki/QA:Updates_Testing for more '
+                'information on how to test updates.'),
+            'validator': str},
         'testing_bug_msg': {
             'value': (
-                '\nSee https://fedoraproject.org/wiki/QA:Updates_Testing for\ninstructions on how '
-                'to install test updates.\nYou can provide feedback for this update here: %s'),
-            'validator': six.text_type},
+                '{update_alias} has been pushed to the {repo} repository.\n'
+                '{install_instructions}\n'
+                'You can provide feedback for this update here: {update_url}\n\n'
+                'See also https://fedoraproject.org/wiki/QA:Updates_Testing for more '
+                'information on how to test updates.'),
+            'validator': str},
         'top_testers_timeframe': {
             'value': 7,
             'validator': int},
@@ -638,48 +601,51 @@ class BodhiConfig(dict):
             'validator': _validate_bool},
         'test_gating.url': {
             'value': '',
-            'validator': six.text_type},
+            'validator': str},
         'updateinfo_rights': {
             'value': 'Copyright (C) {} Red Hat, Inc. and others.'.format(datetime.now().year),
-            'validator': six.text_type},
+            'validator': str},
         'wait_for_repo_sig': {
             'value': False,
             'validator': _validate_bool},
+        'warm_cache_on_start': {
+            'value': True,
+            'validator': _validate_bool},
         'wiki_url': {
             'value': 'https://fedoraproject.org/w/api.php',
-            'validator': six.text_type},
+            'validator': str},
     }
 
-    def __getitem__(self, *args, **kw):
+    def __getitem__(self, key: typing.Hashable) -> typing.Any:
         """Ensure the config is loaded, and then call the superclass __getitem__."""
         if not self.loaded:
             self.load_config()
-        return super(BodhiConfig, self).__getitem__(*args, **kw)
+        return super(BodhiConfig, self).__getitem__(key)
 
-    def get(self, *args, **kw):
+    def get(self, *args, **kw) -> typing.Any:
         """Ensure the config is loaded, and then call the superclass get."""
         if not self.loaded:
             self.load_config()
         return super(BodhiConfig, self).get(*args, **kw)
 
-    def pop(self, *args, **kw):
+    def pop(self, *args, **kw) -> typing.Any:
         """Ensure the config is loaded, and then call the superclass pop."""
         if not self.loaded:
             self.load_config()
         return super(BodhiConfig, self).pop(*args, **kw)
 
-    def copy(self, *args, **kw):
+    def copy(self) -> typing.Any:
         """Ensure the config is loaded, and then call the superclass copy."""
         if not self.loaded:
             self.load_config()
-        return super(BodhiConfig, self).copy(*args, **kw)
+        return super(BodhiConfig, self).copy()
 
-    def load_config(self, settings=None):
+    def load_config(self, settings: typing.Mapping = None):
         """
         Load the configuration either from the config file, or from the given settings.
 
         args:
-            settings (dict): If given, the settings are pulled from this dictionary. Otherwise, the
+            settings: If given, the settings are pulled from this dictionary. Otherwise, the
                 config file is used.
         """
         self._load_defaults()
@@ -691,6 +657,11 @@ class BodhiConfig(dict):
         self.loaded = True
         self._validate()
 
+    def clear(self):
+        """Restore the configuration to its original blank state."""
+        super().clear()
+        self.loaded = False
+
     def _load_defaults(self):
         """Iterate over self._defaults and set all default values on self."""
         for k, v in self._defaults.items():
@@ -703,7 +674,7 @@ class BodhiConfig(dict):
             try:
                 self[k] = self._defaults[k]['validator'](self[k])
             except ValueError as e:
-                errors.append('\t{}: {}'.format(k, six.text_type(e)))
+                errors.append('\t{}: {}'.format(k, str(e)))
 
         if errors:
             raise ValueError(
