@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 # Copyright © 2007-2017 Red Hat, Inc. and others.
 #
 # This file is part of Bodhi.
@@ -29,3 +28,7 @@ class RepodataException(Exception):
 
 class LockedUpdateException(Exception):
     """Raised when something attempts to operate on a locked update."""
+
+
+class ExternalCallException(Exception):
+    """Raised when a call to an external service fails."""
