@@ -1,5 +1,4 @@
-# -*- coding: utf-8 -*-
-# Copyright © 2013-2017 Red Hat, Inc. and others.
+# Copyright © 2013-2019 Red Hat, Inc. and others.
 #
 # This file is part of Bodhi.
 #
@@ -17,7 +16,6 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 """A set of API schemas to validate input and generate documentation."""
-import re
 import os
 
 import colander
@@ -26,7 +24,9 @@ from bodhi.server import util
 from bodhi.server.config import config
 from bodhi.server.models import (
     ContentType,
+    PackageManager,
     ReleaseState,
+    TestGatingStatus,
     UpdateRequest,
     UpdateSeverity,
     UpdateStatus,
@@ -36,8 +36,6 @@ from bodhi.server.models import (
 from bodhi.server.validators import validate_csrf_token
 
 
-CVE_REGEX = re.compile(r"CVE-[0-9]{4,4}-[0-9]{4,}")
-
 # Retrieving list of templates from filesystem for `mail_template` validation in SaveReleaseSchema
 template_directory = util.get_absolute_path(config.get('mail.templates_basepath'))
 MAIL_TEMPLATES = [os.path.splitext(file)[0] for file in os.listdir(template_directory)]
@@ -65,25 +63,6 @@ class Builds(colander.SequenceSchema):
     build = colander.SchemaNode(colander.String())
 
 
-class CVE(colander.String):
-    """A String schema to validate a CVE."""
-
-    def deserialize(self, node, cstruct):
-        """Parse a CVE out of a given API CVE parameter."""
-        value = super(CVE, self).deserialize(node, cstruct)
-
-        if CVE_REGEX.match(value) is None:
-            raise colander.Invalid(node, '"%s" is not a valid CVE id' % value)
-
-        return value
-
-
-class CVEs(colander.SequenceSchema):
-    """A SequenceSchema to validate a list of CVE objects."""
-
-    cve = colander.SchemaNode(CVE())
-
-
 class Packages(colander.SequenceSchema):
     """A SequenceSchema to validate a list of Package objects."""
 
@@ -120,6 +99,13 @@ class Updates(colander.SequenceSchema):
     update = colander.SchemaNode(colander.String())
 
 
+class Status(colander.SequenceSchema):
+    """A SequenceSchema to validate a list of Update status objects."""
+
+    status = colander.SchemaNode(colander.String(),
+                                 validator=colander.OneOf(list(UpdateStatus.values())))
+
+
 class Tests(colander.SequenceSchema):
     """A SequenceSchema to validate a list of Test objects."""
 
@@ -186,15 +172,6 @@ class SaveCommentSchema(CSRFProtectedSchema, colander.MappingSchema):
     bug_feedback = BugFeedbacks(missing=[])
     testcase_feedback = TestcaseFeedbacks(missing=[])
 
-    # Optional
-    captcha_key = colander.SchemaNode(colander.String(), missing=None)
-    captcha_value = colander.SchemaNode(colander.String(), missing=None)
-    email = colander.SchemaNode(
-        colander.String(),
-        validator=colander.Email(),
-        missing=None,
-    )
-
 
 class SaveUpdateSchema(CSRFProtectedSchema, colander.MappingSchema):
     """An API schema for bodhi.server.services.updates.new_update()."""
@@ -202,8 +179,17 @@ class SaveUpdateSchema(CSRFProtectedSchema, colander.MappingSchema):
     builds = Builds(colander.Sequence(accept_scalar=True),
                     preparer=[util.splitter])
 
+    from_tag = colander.SchemaNode(
+        colander.String(),
+        missing=None,
+    )
+
     bugs = Bugs(colander.Sequence(accept_scalar=True), missing=None, preparer=[util.splitter])
 
+    display_name = colander.SchemaNode(
+        colander.String(),
+        missing='',
+    )
     close_bugs = colander.SchemaNode(
         colander.Boolean(),
         missing=True,
@@ -212,11 +198,6 @@ class SaveUpdateSchema(CSRFProtectedSchema, colander.MappingSchema):
         colander.String(),
         validator=colander.OneOf(list(UpdateType.values())),
     )
-    request = colander.SchemaNode(
-        colander.String(),
-        validator=colander.OneOf(list(UpdateRequest.values())),
-        missing='testing',
-    )
     severity = colander.SchemaNode(
         colander.String(),
         validator=colander.OneOf(list(UpdateSeverity.values())),
@@ -262,6 +243,15 @@ class SaveUpdateSchema(CSRFProtectedSchema, colander.MappingSchema):
         colander.Boolean(),
         missing=True,
     )
+    autotime = colander.SchemaNode(
+        colander.Boolean(),
+        missing=True,
+    )
+    stable_days = colander.SchemaNode(
+        colander.Integer(),
+        validator=colander.Range(min=0),
+        missing=0,
+    )
 
 
 class Cosmetics(colander.MappingSchema):
@@ -409,6 +399,7 @@ class SaveReleaseSchema(CSRFProtectedSchema, colander.MappingSchema):
     )
     override_tag = colander.SchemaNode(
         colander.String(),
+        missing="",
     )
     state = colander.SchemaNode(
         colander.String(),
@@ -428,45 +419,21 @@ class SaveReleaseSchema(CSRFProtectedSchema, colander.MappingSchema):
         colander.Boolean(true_choices=('true', '1')),
         missing=True,
     )
-
-
-class ListStackSchema(PaginatedSchema, SearchableSchema):
-    """An API schema for bodhi.server.services.stacks.query_stacks()."""
-
-    name = colander.SchemaNode(
+    create_automatic_updates = colander.SchemaNode(
+        colander.Boolean(true_choices=('true', '1')),
+        missing=False,
+    )
+    package_manager = colander.SchemaNode(
         colander.String(),
-        location="querystring",
-        missing=None,
+        validator=colander.OneOf(list(PackageManager.values())),
+        missing="unspecified",
     )
-
-    packages = Packages(
-        colander.Sequence(accept_scalar=True),
-        location="querystring",
-        missing=None,
-        preparer=[util.splitter],
-    )
-
-
-class SaveStackSchema(CSRFProtectedSchema, colander.MappingSchema):
-    """An API schema for bodhi.server.services.stacks.save_stack()."""
-
-    name = colander.SchemaNode(
-        colander.String(),
-    )
-
-    packages = Packages(
-        colander.Sequence(accept_scalar=True),
-        missing=None,
-        preparer=[util.splitter],
-    )
-
-    description = colander.SchemaNode(
+    testing_repository = colander.SchemaNode(
         colander.String(),
         missing=None,
     )
-
-    requirements = colander.SchemaNode(
-        colander.String(),
+    eol = colander.SchemaNode(
+        colander.Date(),
         missing=None,
     )
 
@@ -544,11 +511,10 @@ class ListUpdateSchema(PaginatedSchema, SearchableSchema, Cosmetics):
         missing=None,
     )
 
-    cves = CVEs(
-        colander.Sequence(accept_scalar=True),
+    from_side_tag = colander.SchemaNode(
+        colander.Boolean(true_choices=('true', '1')),
         location="querystring",
         missing=None,
-        preparer=[util.splitter],
     )
 
     locked = colander.SchemaNode(
@@ -569,12 +535,6 @@ class ListUpdateSchema(PaginatedSchema, SearchableSchema, Cosmetics):
         missing=None,
     )
 
-    active_releases = Releases(
-        colander.Boolean(true_choices=('true', '1')),
-        location="querystring",
-        missing=False,
-    )
-
     packages = Packages(
         colander.Sequence(accept_scalar=True),
         location="querystring",
@@ -629,11 +589,11 @@ class ListUpdateSchema(PaginatedSchema, SearchableSchema, Cosmetics):
         validator=colander.OneOf(list(UpdateSeverity.values())),
     )
 
-    status = colander.SchemaNode(
-        colander.String(),
+    status = Status(
+        colander.Sequence(accept_scalar=True),
         location="querystring",
         missing=None,
-        validator=colander.OneOf(list(UpdateStatus.values())),
+        preparer=[util.splitter],
     )
 
     submitted_since = colander.SchemaNode(
@@ -683,6 +643,13 @@ class ListUpdateSchema(PaginatedSchema, SearchableSchema, Cosmetics):
         preparer=[util.splitter],
     )
 
+    gating = colander.SchemaNode(
+        colander.String(),
+        location="querystring",
+        missing=None,
+        validator=colander.OneOf(list(TestGatingStatus.values())),
+    )
+
 
 class ListPackageSchema(PaginatedSchema, SearchableSchema):
     """An API schema for bodhi.server.services.packages.query_packages()."""
@@ -772,12 +739,6 @@ class ListCommentSchema(PaginatedSchema, SearchableSchema):
         preparer=[util.splitter],
     )
 
-    anonymous = colander.SchemaNode(
-        colander.Boolean(true_choices=('true', '1')),
-        location="querystring",
-        missing=None,
-    )
-
     since = colander.SchemaNode(
         colander.DateTime(),
         location="querystring",
@@ -867,3 +828,9 @@ class GetTestResultsSchema(CSRFProtectedSchema, colander.MappingSchema):
         colander.Sequence(accept_scalar=True),
         missing=None,
     )
+
+
+class TriggerTestsSchema(CSRFProtectedSchema, colander.MappingSchema):
+    """An API schema for bodhi.server.services.updates.trigger_tests()."""
+
+    pass
