@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 # Copyright © 2018 Red Hat, Inc.
 #
 # This file is part of Bodhi.
@@ -37,6 +36,7 @@ Some other things that aren't implemented (but could be added if necessary):
  - Handling foreign layers
 """
 
+from urllib.parse import urlparse, urlunparse
 import json
 import logging
 import os
@@ -46,7 +46,6 @@ import tempfile
 import click
 import requests
 from requests.exceptions import SSLError, ConnectionError
-from six.moves.urllib.parse import urlparse, urlunparse
 
 
 @click.group()
@@ -204,16 +203,16 @@ class RegistrySession(object):
         if cert_dir is None:
             return None
 
-        for l in sorted(os.listdir(cert_dir)):
-            if l.endswith('.cert'):
-                certpath = os.path.join(cert_dir, l)
+        for d in sorted(os.listdir(cert_dir)):
+            if d.endswith('.cert'):
+                certpath = os.path.join(cert_dir, d)
                 keypath = certpath[:-5] + '.key'
                 if not os.path.exists(keypath):
                     raise RuntimeError("Cannot find key file for {}".format(certpath))
                 return (certpath, keypath)
-            elif l.endswith('.key'):
+            elif d.endswith('.key'):
                 # Should have found <x>.cert first
-                keypath = os.path.join(cert_dir, l)
+                keypath = os.path.join(cert_dir, d)
                 raise RuntimeError("Cannot find certificate file for {}".format(keypath))
 
         return None
@@ -690,9 +689,14 @@ class Copier(object):
             self.src.download_blob(digest, size, self.dest.ensure_blob_path(digest))
         elif isinstance(self.src, DirectoryEndpoint) and isinstance(self.dest, RegistryEndpoint):
             self.dest.upload_blob(digest, size, self.src.get_blob_path(digest))
-        elif isinstance(self.src, RegistryEndpoint) and isinstance(self.dest, RegistryEndpoint):
-            if self.src.registry == self.dest.registry:
-                self.dest.link_blob(digest, self.src.repo)
+        else:
+            # The copy() function below ensures that the following two asserts don't fail.
+            assert (isinstance(self.src, RegistryEndpoint)
+                    and isinstance(self.dest, RegistryEndpoint))
+            assert self.src.registry == self.dest.registry
+
+            self.dest.link_blob(digest, self.src.repo)
+
         # Other forms of copying are not needed currently, and not implemented
 
     def _copy_manifest(self, info, toplevel=False):
@@ -762,6 +766,11 @@ def copy(src, dest, src_creds, src_tls_verify, src_cert_dir, dest_creds, dest_tl
     src = parse_spec(src, src_creds, src_tls_verify, src_cert_dir)
     dest = parse_spec(dest, dest_creds, dest_tls_verify, dest_cert_dir)
 
+    # Copier._copy_blob() assumes this:
+    # - The src and dest registries are of types RegistryEndpoint or DirectoryEndpoint, at least one
+    #   must be a RegistryEndpoint.
+    # - If both are a RegistryEndpoint, they must be the same so linking is possible, therefore we
+    #   copy src reg -> local dir -> dest reg if this isn't the case.
     if src.registry != dest.registry:
         tempdir = tempfile.mkdtemp()
         try:
