@@ -1,5 +1,4 @@
-# -*- coding: utf-8 -*-
-# Copyright © 2014-2017 Red Hat, Inc. and others.
+# Copyright © 2014-2019 Red Hat, Inc. and others.
 #
 # This file is part of Bodhi.
 #
@@ -23,43 +22,47 @@ import math
 
 from cornice import Service
 from cornice.validators import colander_body_validator, colander_querystring_validator
-from sqlalchemy import func, distinct
+from requests import RequestException
+from requests import Timeout as RequestsTimeout
+from sqlalchemy import distinct, func
 from sqlalchemy.sql import or_
-from requests import RequestException, Timeout as RequestsTimeout
-import six
 
+from bodhi.messages.schemas import update as update_schemas
 from bodhi.server import log, security
 from bodhi.server.exceptions import BodhiException, LockedUpdateException
 from bodhi.server.models import (
-    Update,
     Bug,
-    ContentType,
-    CVE,
-    UpdateRequest,
-    Release,
-    ReleaseState,
     Build,
+    ContentType,
     Package,
+    Release,
+    Update,
+    UpdateRequest,
+    UpdateStatus,
 )
+from bodhi.server.tasks import handle_side_and_related_tags_task
+from bodhi.server.validators import (
+    validate_acls,
+    validate_bugs,
+    validate_build_nvrs,
+    validate_build_tags,
+    validate_build_uniqueness,
+    validate_builds,
+    validate_builds_or_from_tag_exist,
+    validate_enums,
+    validate_from_tag,
+    validate_release,
+    validate_releases,
+    validate_request,
+    validate_requirements,
+    validate_severity,
+    validate_update_id,
+    validate_username,
+)
+import bodhi.server.notifications as notifications
 import bodhi.server.schemas
 import bodhi.server.services.errors
 import bodhi.server.util
-from bodhi.server.validators import (
-    validate_nvrs,
-    validate_uniqueness,
-    validate_build_tags,
-    validate_acls,
-    validate_builds,
-    validate_enums,
-    validate_releases,
-    validate_release,
-    validate_username,
-    validate_update_id,
-    validate_requirements,
-    validate_bugs,
-    validate_request,
-    validate_severity,
-)
 
 
 update = Service(name='update', path='/updates/{id}',
@@ -103,6 +106,14 @@ update_get_test_results = Service(
     cors_origins=bodhi.server.security.cors_origins_ro,
 )
 
+update_trigger_tests = Service(
+    name='update_trigger_tests',
+    path='/updates/{id}/trigger-tests',
+    description='Trigger tests for a specific update',
+    factory=security.PackagerACLFactory,
+    cors_origins=bodhi.server.security.cors_origins_rw
+)
+
 
 @update.get(accept=('application/json', 'text/json'), renderer='json',
             error_handler=bodhi.server.services.errors.json_handler)
@@ -112,7 +123,7 @@ update_get_test_results = Service(
             error_handler=bodhi.server.services.errors.html_handler)
 def get_update(request):
     """
-    Return a single update from an id, title, or alias.
+    Return a single update from an id or alias.
 
     Args:
         request (pyramid.request): The current request.
@@ -137,7 +148,7 @@ def get_update(request):
                  ))
 def get_update_for_editing(request):
     """
-    Return a single update from an id, title, or alias for the edit form.
+    Return a single update from an id or alias for the edit form.
 
     Args:
         request (pyramid.request): The current request.
@@ -147,30 +158,33 @@ def get_update_for_editing(request):
             types: The possible values for update types.
             severities: The possible values for update severity.
             suggestions: The possible values for update suggestion.
+            sidetags: a list of sidetags for the user, always empty
+            in the edit form case.
     """
     suggestions = list(bodhi.server.models.UpdateSuggestion.values())
-    if six.PY2:  # pragma: no cover
-        suggestions = reversed(suggestions)
     return dict(
         update=request.validated['update'],
         types=reversed(list(bodhi.server.models.UpdateType.values())),
         severities=sorted(
             list(bodhi.server.models.UpdateSeverity.values()), key=bodhi.server.util.sort_severity),
         suggestions=suggestions,
+        sidetags=[],
     )
 
 
-@update_request.post(schema=bodhi.server.schemas.UpdateRequestSchema,
-                     validators=(
-                         colander_body_validator,
-                         validate_enums,
-                         validate_update_id,
-                         validate_build_tags,
-                         validate_acls,
-                         validate_request,
-                     ),
-                     permission='edit', renderer='json',
-                     error_handler=bodhi.server.services.errors.json_handler)
+@update_request.post(
+    schema=bodhi.server.schemas.UpdateRequestSchema(),
+    validators=(
+        colander_body_validator,
+        validate_enums,
+        validate_update_id,
+        validate_build_tags,
+        validate_acls,
+        validate_request,
+    ),
+    permission='edit', renderer='json',
+    error_handler=bodhi.server.services.errors.json_handler
+)
 def set_request(request):
     """
     Set a specific :class:`bodhi.server.models.UpdateRequest` on a given update.
@@ -188,15 +202,18 @@ def set_request(request):
                            "Can't change request on a locked update")
         return
 
-    if update.release.state is ReleaseState.archived:
+    if update.status == UpdateStatus.stable and action == UpdateRequest.testing:
         request.errors.add('body', 'request',
-                           "Can't change request for an archived release")
+                           "Pushing back to testing a stable update is not allowed")
         return
 
-    if action in (UpdateRequest.stable, UpdateRequest.batched):
+    if action == UpdateRequest.stable:
         settings = request.registry.settings
         result, reason = update.check_requirements(request.db, settings)
         if not result:
+            log.info(
+                f'Unable to set request for {update.alias} to stable due to failed requirements: '
+                f'{reason}')
             request.errors.add('body', 'request',
                                'Requirement not met %s' % reason)
             return
@@ -204,7 +221,7 @@ def set_request(request):
     try:
         update.set_request(request.db, action, request.user.name)
     except BodhiException as e:
-        log.error("Failed to set the request: %s", e)
+        log.info("Failed to set the request: %s", e)
         request.errors.add('body', 'request', str(e))
     except Exception as e:
         log.exception("Unhandled exception in set_request")
@@ -223,22 +240,22 @@ validators = (
 )
 
 
-@updates_rss.get(schema=bodhi.server.schemas.ListUpdateSchema, renderer='rss',
+@updates_rss.get(schema=bodhi.server.schemas.ListUpdateSchema(), renderer='rss',
                  error_handler=bodhi.server.services.errors.html_handler,
                  validators=validators)
-@updates.get(schema=bodhi.server.schemas.ListUpdateSchema, renderer='rss',
+@updates.get(schema=bodhi.server.schemas.ListUpdateSchema(), renderer='rss',
              accept=('application/atom+xml',),
              error_handler=bodhi.server.services.errors.html_handler,
              validators=validators)
-@updates.get(schema=bodhi.server.schemas.ListUpdateSchema,
+@updates.get(schema=bodhi.server.schemas.ListUpdateSchema(),
              accept=('application/json', 'text/json'), renderer='json',
              error_handler=bodhi.server.services.errors.json_handler,
              validators=validators)
-@updates.get(schema=bodhi.server.schemas.ListUpdateSchema,
+@updates.get(schema=bodhi.server.schemas.ListUpdateSchema(),
              accept=('application/javascript'), renderer='jsonp',
              error_handler=bodhi.server.services.errors.jsonp_handler,
              validators=validators)
-@updates.get(schema=bodhi.server.schemas.ListUpdateSchema,
+@updates.get(schema=bodhi.server.schemas.ListUpdateSchema(),
              accept=('text/html'), renderer='updates.html',
              error_handler=bodhi.server.services.errors.html_handler,
              validators=validators)
@@ -278,21 +295,16 @@ def query_updates(request):
     if critpath is not None:
         query = query.filter(Update.critpath == critpath)
 
-    cves = data.get('cves')
-    if cves is not None:
-        query = query.join(Update.cves)
-        query = query.filter(or_(*[CVE.cve_id == cve_id for cve_id in cves]))
-
     like = data.get('like')
     if like is not None:
-        query = query.filter(or_(*[
-            Update.title.like('%%%s%%' % like)
-        ]))
+        query = query.join(Update.builds)
+        query = query.filter(Build.nvr.like('%%%s%%' % like))
 
     search = data.get('search')
     if search is not None:
-        query = query.filter(or_(Update.title.ilike('%%%s%%' % search),
-                                 Update.alias.ilike('%%%s%%' % search)))
+        query = query.join(Update.builds)
+        query = query.filter(or_(
+            Build.nvr.ilike('%%%s%%' % search), Update.alias.ilike('%%%s%%' % search)))
 
     locked = data.get('locked')
     if locked is not None:
@@ -333,15 +345,8 @@ def query_updates(request):
         query = query.filter(Update.date_pushed < pushed_before)
 
     releases = data.get('releases')
-    active_releases = data.get('active_releases')
     if releases is not None:
         query = query.filter(or_(*[Update.release == r for r in releases]))
-    elif active_releases:
-        query = query.filter(
-            Update.release_id == Release.id
-        ).filter(
-            Release.state.in_([ReleaseState.current, ReleaseState.pending])
-        )
 
     # This singular version of the plural "releases" is purely for bodhi1
     # backwards compat (mostly for RSS feeds) - threebean
@@ -359,7 +364,7 @@ def query_updates(request):
 
     status = data.get('status')
     if status is not None:
-        query = query.filter(Update.status == status)
+        query = query.filter(or_(*[Update.status == s for s in status]))
 
     submitted_since = data.get('submitted_since')
     if submitted_since is not None:
@@ -382,6 +387,10 @@ def query_updates(request):
         query = query.join(Update.builds)
         query = query.filter(Build.type == content_type)
 
+    gating_status = data.get('gating')
+    if gating_status is not None:
+        query = query.filter(Update.test_gating_status == gating_status)
+
     user = data.get('user')
     if user is not None:
         query = query.filter(or_(*[Update.user == u for u in user]))
@@ -389,10 +398,18 @@ def query_updates(request):
     updateid = data.get('updateid')
     if updateid is not None:
         query = query.filter(or_(*[Update.alias == uid for uid in updateid]))
+
     alias = data.get('alias')
     if alias is not None:
         query = query.filter(or_(*[Update.alias == a for a in alias]))
 
+    from_side_tag = data.get('from_side_tag')
+    if from_side_tag is not None:
+        if from_side_tag:
+            query = query.filter(Update.from_tag.isnot(None))
+        else:
+            query = query.filter(Update.from_tag.is_(None))
+
     query = query.order_by(Update.date_submitted.desc())
 
     # We can't use ``query.count()`` here because it is naive with respect to
@@ -407,7 +424,7 @@ def query_updates(request):
     pages = int(math.ceil(total / float(rows_per_page)))
     query = query.offset(rows_per_page * (page - 1)).limit(rows_per_page)
 
-    return dict(
+    return_values = dict(
         updates=query.all(),
         page=page,
         pages=pages,
@@ -417,33 +434,61 @@ def query_updates(request):
         display_user=data.get('display_user', False),
         display_request=data.get('display_request', True),
         package=package,
-        active_releases=active_releases,
     )
+    # we need some extra information for the searching / filterings interface
+    # when rendering the html, so we add this here.
+    if request.accept.accept_html():
+        return_values.update(
+            gating_statuses=sorted(
+                list(bodhi.server.models.TestGatingStatus.values())),
+            types=list(bodhi.server.models.UpdateType.values()),
+            severities=sorted(
+                list(bodhi.server.models.UpdateSeverity.values()),
+                key=bodhi.server.util.sort_severity),
+            statuses=list(bodhi.server.models.UpdateStatus.values()),
+            releases=Release.all_releases(),
+        )
+    return return_values
 
 
-@updates.post(schema=bodhi.server.schemas.SaveUpdateSchema,
-              permission='create', renderer='json',
-              error_handler=bodhi.server.services.errors.json_handler,
-              validators=(
-                  colander_body_validator,
-                  validate_nvrs,
-                  validate_builds,
-                  validate_build_tags,
-                  validate_uniqueness,
-                  validate_acls,
-                  validate_enums,
-                  validate_requirements,
-                  validate_bugs,
-                  validate_severity,
-              ))
+@updates.post(
+    schema=bodhi.server.schemas.SaveUpdateSchema(),
+    permission='create', renderer='json',
+    error_handler=bodhi.server.services.errors.json_handler,
+    validators=(
+        colander_body_validator,
+        validate_from_tag,
+        validate_build_nvrs,
+        validate_builds,
+        validate_build_tags,
+        validate_build_uniqueness,
+        validate_builds_or_from_tag_exist,
+        validate_acls,
+        validate_enums,
+        validate_requirements,
+        validate_bugs,
+        validate_severity,
+    )
+)
 def new_update(request):
     """
     Save an update.
 
     This entails either creating a new update, or editing an existing one. To
-    edit an existing update, the update's original title must be specified in
+    edit an existing update, the update's alias must be specified in
     the ``edited`` parameter.
 
+    If the ``from_tag`` parameter is specified and ``builds`` is missing or
+    empty, the list of builds will be filled with the latest builds in this
+    Koji tag. This is done by validate_from_tag() because the list of builds
+    needs to be available in validate_acls().
+
+    If the release is composed by Bodhi (i.e. a branched or stable release
+    after the Bodhi activation point), ensure that related tags
+    ``from_tag``-pending-signing and ``from_tag``-testing exists and if not
+    create them in Koji. If the state of the release is not `pending`, add its
+    pending-signing tag and remove it if it's a side tag.
+
     Args:
         request (pyramid.request): The current request.
     """
@@ -454,6 +499,13 @@ def new_update(request):
     # it since the models don't care about a csrf argument.
     data.pop('csrf_token')
 
+    # Same here, but it can be missing.
+    data.pop('builds_from_tag', None)
+    data.pop('sidetag_owner', None)
+
+    build_nvrs = data.get('builds', [])
+    from_tag = data.get('from_tag')
+
     caveats = []
     try:
 
@@ -461,10 +513,10 @@ def new_update(request):
         builds = []
 
         # Create the Package and Build entities
-        for nvr in data['builds']:
+        for nvr in build_nvrs:
             name, version, release = request.buildinfo[nvr]['nvr']
 
-            package = Package.get_or_create(request.buildinfo[nvr])
+            package = Package.get_or_create(request.db, request.buildinfo[nvr])
 
             # Also figure out the build type and create the build if absent.
             build_class = ContentType.infer_content_class(
@@ -482,6 +534,18 @@ def new_update(request):
             builds.append(build)
             releases.add(request.buildinfo[build.nvr]['release'])
 
+        # Disable manual updates for releases not composed by Bodhi
+        # see #4058
+        if not from_tag:
+            for release in releases:
+                if not release.composed_by_bodhi:
+                    request.errors.add('body', 'builds',
+                                       "Cannot manually create updates for a Release which is not "
+                                       "composed by Bodhi.\nRead the 'Automatic updates' page in "
+                                       "Bodhi docs about this error.")
+                    request.db.rollback()
+                    return
+
         # We want to go ahead and commit the transaction now so that the Builds are in the database.
         # Otherwise, there will be a race condition between robosignatory signing the Builds and the
         # signed handler attempting to mark the builds as signed. When we lose that race, the signed
@@ -489,11 +553,11 @@ def new_update(request):
         # the builds as signed.
         request.db.commit()
 
-        # After we commit the transaction, we need to get the builds and releases again, since they
-        # were tied to the previous session that has now been terminated.
+        # After we commit the transaction, we need to get the builds and releases again,
+        # since they were tied to the previous session that has now been terminated.
         builds = []
         releases = set()
-        for nvr in data['builds']:
+        for nvr in build_nvrs:
             # At this moment, we are sure the builds are in the database (that is what the commit
             # was for actually).
             build = Build.get(nvr)
@@ -504,9 +568,9 @@ def new_update(request):
 
             log.info('Editing update: %s' % data['edited'])
 
-            assert len(releases) == 1, "Updates may not span multiple releases"
             data['release'] = list(releases)[0]
             data['builds'] = [b.nvr for b in builds]
+            data['from_tag'] = from_tag
             result, _caveats = Update.edit(request, data)
             caveats.extend(_caveats)
         else:
@@ -522,16 +586,43 @@ def new_update(request):
                 _data = copy.copy(data)  # Copy it because .new(..) mutates it
                 _data['builds'] = [b for b in builds if b.release == release]
                 _data['release'] = release
+                _data['from_tag'] = from_tag
 
                 log.info('Creating new update: %r' % _data['builds'])
                 result, _caveats = Update.new(request, _data)
-                log.debug('%s update created', result.title)
+                log.debug('%s update created', result.alias)
 
                 updates.append(result)
                 caveats.extend(_caveats)
 
             if len(releases) > 1:
                 result = dict(updates=updates)
+
+            if from_tag:
+                for u in updates:
+                    builds = [b.nvr for b in u.builds]
+                    if not u.release.composed_by_bodhi:
+                        # Before the Bodhi activation point of a release, keep builds tagged
+                        # with the side-tag and its associate tags.
+                        side_tag_signing_pending = u.release.get_pending_signing_side_tag(from_tag)
+                        side_tag_testing_pending = u.release.get_pending_testing_side_tag(from_tag)
+                        handle_side_and_related_tags_task.delay(
+                            builds=builds,
+                            pending_signing_tag=side_tag_signing_pending,
+                            from_tag=from_tag,
+                            pending_testing_tag=side_tag_testing_pending)
+                    else:
+                        # After the Bodhi activation point of a release, add the pending-signing tag
+                        # of the release to funnel the builds back into a normal workflow for a
+                        # stable release.
+                        pending_signing_tag = u.release.pending_signing_tag
+                        candidate_tag = u.release.candidate_tag
+                        handle_side_and_related_tags_task.delay(
+                            builds=builds,
+                            pending_signing_tag=pending_signing_tag,
+                            from_tag=from_tag,
+                            candidate_tag=candidate_tag)
+
     except LockedUpdateException as e:
         log.warning(str(e))
         request.errors.add('body', 'builds', "%s" % str(e))
@@ -567,7 +658,7 @@ def new_update(request):
     return result
 
 
-@update_waive_test_results.post(schema=bodhi.server.schemas.WaiveTestResultsSchema,
+@update_waive_test_results.post(schema=bodhi.server.schemas.WaiveTestResultsSchema(),
                                 validators=(colander_body_validator,
                                             validate_update_id,
                                             validate_acls),
@@ -601,7 +692,7 @@ def waive_test_results(request):
     return dict(update=update)
 
 
-@update_get_test_results.get(schema=bodhi.server.schemas.GetTestResultsSchema,
+@update_get_test_results.get(schema=bodhi.server.schemas.GetTestResultsSchema(),
                              validators=(validate_update_id),
                              renderer='json',
                              error_handler=bodhi.server.services.errors.json_handler)
@@ -612,13 +703,13 @@ def get_test_results(request):
     Args:
         request (pyramid.request): The current request.
     Returns:
-        dict: A dictionary mapping the key "update" to the update.
+        dict: A dictionary mapping the key 'decisions' to a list of result dictionaries.
     """
     update = request.validated['update']
 
-    decision = None
+    decisions = []
     try:
-        decision = update.get_test_gating_info()
+        decisions = update.get_test_gating_info()
     except RequestsTimeout as e:
         log.error("Error querying greenwave for test results - timed out")
         request.errors.add('body', 'request', str(e))
@@ -636,4 +727,34 @@ def get_test_results(request):
         request.errors.add('body', 'request', str(e))
         request.errors.status = 500
 
-    return dict(decision=decision)
+    return dict(decisions=decisions)
+
+
+@update_trigger_tests.post(schema=bodhi.server.schemas.TriggerTestsSchema(),
+                           validators=(colander_body_validator,
+                                       validate_update_id,
+                                       validate_acls),
+                           permission='edit', renderer='json',
+                           error_handler=bodhi.server.services.errors.json_handler)
+def trigger_tests(request):
+    """
+    Trigger tests for update.
+
+    Args:
+        request (pyramid.request): The current request.
+    Returns:
+        dict: A dictionary mapping the key "update" to the update.
+    """
+    update = request.validated['update']
+
+    if update.status != UpdateStatus.testing:
+        log.error("Can't trigger tests for update: Update is not in testing status")
+        request.errors.add('body', 'request', 'Update is not in testing status')
+    else:
+        if update.content_type == ContentType.rpm:
+            message = update_schemas.UpdateReadyForTestingV2.from_dict(
+                message=update._build_group_test_message(agent=request.user.name, retrigger=True)
+            )
+            notifications.publish(message)
+
+    return dict(update=update)
