@@ -1,5 +1,3 @@
-var cabbage = new Cabbage();
-
 var delay = (function(){
   var timer = 0;
   return function(callback, ms){
@@ -16,7 +14,6 @@ var update_markdown_preview = function(text) {
             data: $.param({text: text}),
             dataType: 'json',
             success: function(data) {
-                console.log(data)
                 $("#preview").html(data.html);
             },
             error: function(e1, e2, e3) {
@@ -40,6 +37,8 @@ $(document).ready(function() {
 
     // Attach bootstrap tooltips to everything.
     $('[data-toggle="tooltip"]').tooltip();
+    $('[data-toggle="popover"]').popover();
+
 
     // Make the rows on the comment form change color on click.
     $('.table td > input').click(function() {
@@ -49,14 +48,4 @@ $(document).ready(function() {
         td.parent().addClass(td.attr('data-class'));
     });
 
-    $('.searchbar .typeahead').blur(function(){
-        $('.searchbar').slideUp();
-    });
-
-    $('#search-toggle').click(function(){
-      if ($('.searchbar').is(":hidden")){
-        $('.searchbar').slideDown();
-        $('.searchbar .typeahead').focus();
-      }
-    });
 });
