@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 # Copyright © 2014-2019 Red Hat, Inc. and others
 #
 # This file is part of Bodhi.
@@ -20,15 +19,16 @@
 
 import datetime
 
+from prometheus_client import CONTENT_TYPE_LATEST, generate_latest, REGISTRY
+from pyramid.exceptions import HTTPBadRequest, HTTPForbidden
+from pyramid.httpexceptions import HTTPMovedPermanently, HTTPUnauthorized
+from pyramid.response import Response
 from pyramid.settings import asbool
-from pyramid.view import view_config, notfound_view_config
-from pyramid.exceptions import HTTPForbidden
-from pyramid.httpexceptions import HTTPFound
+from pyramid.view import notfound_view_config, view_config
 import cornice.errors
 import sqlalchemy as sa
-import six
 
-from bodhi.server import log, models
+from bodhi.server import log, METADATA, models
 from bodhi.server.config import config
 import bodhi.server.util
 
@@ -49,14 +49,14 @@ def get_top_testers():
 
     query = models.Session().query(
         models.User,
-        sa.func.count(models.User.comments).label(u'count_1')
+        sa.func.count(models.User.comments).label('count_1')
     ).join(models.Comment)
     query = query\
-        .order_by(sa.text(u'count_1 desc'))\
+        .order_by(sa.text('count_1 desc'))\
         .filter(models.Comment.timestamp > start_time)
 
     for user in blacklist:
-        query = query.filter(models.User.name != six.text_type(user))
+        query = query.filter(models.User.name != str(user))
 
     return query\
         .group_by(models.User)\
@@ -64,9 +64,40 @@ def get_top_testers():
         .all()
 
 
-def get_latest_updates(critpath, security):
+def get_top_packagers():
     """
-    Return a query of the 5 most recent updates that are in Testing status.
+    Return a query of the 5 users that have submitted the most updates in the last 7 days.
+
+    Returns:
+        sqlalchemy.orm.query.Query: A SQLAlchemy query that contains the
+                                  5 users that have submitted the most updates
+                                  in the last 7 days, and their total number of
+                                  updates in bodhi.
+    """
+    blacklist = config.get('stats_blacklist')
+    days = config.get('top_testers_timeframe')
+    start_time = datetime.datetime.utcnow() - datetime.timedelta(days=days)
+
+    query = models.Session().query(
+        models.User,
+        sa.func.count(models.User.updates).label('count_1')
+    ).join(models.Update)
+    query = query\
+        .order_by(sa.text('count_1 desc'))\
+        .filter(models.Update.date_submitted > start_time)
+
+    for user in blacklist:
+        query = query.filter(models.User.name != str(user))
+
+    return query\
+        .group_by(models.User)\
+        .limit(5)\
+        .all()
+
+
+def get_testing_counts(critpath, security):
+    """
+    Return a count of updates in Testing status.
 
     critpath and security are optional filters for security or critical path updates.
 
@@ -74,8 +105,7 @@ def get_latest_updates(critpath, security):
         critpath (boolean): If True, filter results to only return updates in the Critical Path
         security (boolean): If True, filter results to only return Security updates
     Returns:
-        sqlalchemy.orm.query.Query: A SQLAlchemy query that contains the
-                                  5 most recent updates that are in Testing status
+        int: the number of updates in the testing status
     """
     query = models.Update.query
 
@@ -90,124 +120,128 @@ def get_latest_updates(critpath, security):
         models.Update.status == models.UpdateStatus.testing)
 
     query = query.order_by(models.Update.date_submitted.desc())
-    return query.limit(5).all()
-
-
-def _get_status_counts(basequery, status):
-    """
-    Return a dictionary with the counts of objects found in the basequery.
-
-    The return data is specified by total count, newpackage count, bugfix count, enhancement count,
-    and security count. The dictionary keys will be named with the
-    template {status}_{type}_total. For example, if status is
-    models.UpdateStatus.stable, a dictionary with the following keys would be
-    returned:
-        stable_updates_total
-        stable_newpackage_total
-        stable_bugfix_total
-        stable_enhancement_total
-        stable_security_total
-
-    Args:
-        basequery (sqlalchemy.orm.query.Query):
-            The basequery of updates we want to count and further filter on.
-        status (bodhi.server.models.UpdateStatus):
-            The update status we want to filter by in basequery
-    Return:
-        dict: A dictionary describing the counts of the updates, as described above.
-    """
-    basequery = basequery.filter(models.Update.status == status)
-    return {
-        '{}_updates_total'.format(status.description): basequery.count(),
-        '{}_newpackage_total'.format(status.description):
-            basequery.filter(models.Update.type == models.UpdateType.newpackage).count(),
-        '{}_bugfix_total'.format(status.description):
-            basequery.filter(models.Update.type == models.UpdateType.bugfix).count(),
-        '{}_enhancement_total'.format(status.description):
-            basequery.filter(models.Update.type == models.UpdateType.enhancement).count(),
-        '{}_security_total'.format(status.description):
-            basequery.filter(models.Update.type == models.UpdateType.security).count(),
-    }
-
-
-def get_update_counts(releaseid):
-    """
-    Return counts for the various states and types of updates in the given release.
-
-    This function returns a dictionary that tabulates the counts of the various
-    types of Bodhi updates at the various states they can appear in. The
-    dictionary has the following keys, with pretty self-explanatory names:
-
-        pending_updates_total
-        pending_newpackage_total
-        pending_bugfix_total
-        pending_enhancement_total
-        pending_security_total
-        testing_updates_total
-        testing_newpackage_total
-        testing_bugfix_total
-        testing_enhancement_total
-        testing_security_total
-        stable_updates_total
-        stable_newpackage_total
-        stable_bugfix_total
-        stable_enhancement_total
-        stable_security_total
-
-    Args:
-        request (pyramid.util.Request): The current request
-        releaseid (basestring): The id of the Release object you would like the counts performed on
-    Returns:
-        dict: A dictionary expressing the counts, as described above.
-    """
-    release = models.Release.get(releaseid)
-    basequery = models.Update.query.filter(models.Update.release == release)
-    counts = {}
-    counts.update(_get_status_counts(basequery, models.UpdateStatus.pending))
-    counts.update(_get_status_counts(basequery, models.UpdateStatus.testing))
-    counts.update(_get_status_counts(basequery, models.UpdateStatus.stable))
-
-    return counts
+    return query.count()
 
 
 def _generate_home_page_stats():
     """
     Generate and return a dictionary of stats for the home() function to use.
 
-    This function returns a dictionary with the following 4 keys:
+    This function returns a dictionary with the following 5 keys:
 
-        release_updates_counts: a dict of counts of various states and types of updates as
-                                generated by get_update_counts()
         top_testers:            a list of 5 tuples, the 5 top testers in the last 7 days.
                                 The first item of each tuple contains a dict of the items
                                 in the User model for that user. The second item of the tuple
                                 contains the number of comments the user has left in Bodhi.
-        critpath_updates:       a list of 5 dicts, the 5 most recent critical path updates.
-                                Each dict contains key-value pairs for all the items in the
-                                Update model.
-        security_updates:       a list of 5 dicts, the 5 most recent security updates.
-                                Each dict contains key-value pairs for all the items in the
-                                Update model.
+        top_packagers:           a list of 5 tuples, the 5 top packagers in the last 7 days.
+                                The first item of each tuple contains a dict of the items
+                                in the User model for that user. The second item of the tuple
+                                contains the number of updates the user has filed in Bodhi.
+        critpath_testing_count: the number of critical path updates in testin.
+        security_testing_count: the number of security updates in testing
+        all_testing_count:      the number of all updates in testing
 
     Returns:
         dict: A Dictionary expressing the values described above
     """
     top_testers = get_top_testers()
-    critpath_updates = get_latest_updates(True, False)
-    security_updates = get_latest_updates(False, True)
-    release_updates_counts = {}
-    releases = models.Release.all_releases()
-    for release in releases['current'] + releases['pending']:
-        release_updates_counts[release["name"]] = get_update_counts(release["name"])
+    top_packagers = get_top_packagers()
 
     return {
-        "release_updates_counts": release_updates_counts,
         "top_testers": [(obj.__json__(), n) for obj, n in top_testers],
-        "critpath_updates": [obj.__json__() for obj in critpath_updates],
-        "security_updates": [obj.__json__() for obj in security_updates],
+        "top_packagers": [(obj.__json__(), n) for obj, n in top_packagers],
+        "critpath_testing_count": get_testing_counts(True, False),
+        "security_testing_count": get_testing_counts(False, True),
+        "all_testing_count": get_testing_counts(False, False),
     }
 
 
+def _get_sidetags(koji, user=None, contains_builds=False):
+    """
+    Return a list of koji sidetags.
+
+    Args:
+        koji: the koji client instance from request.koji
+        user: a string of the user's FAS name
+        contains_builds: a boolean. if true, only return sidetags with builds
+    Returns:
+        dict: A list of the sidetags information.
+    """
+    sidetags = koji.listSideTags(user=user)
+
+    koji.multicall = True
+    for tag in sidetags:
+        koji.getTag(tag['name'])
+    sidetags_info = koji.multiCall()
+
+    koji.multicall = True
+    for tag in sidetags_info:
+        koji.listTagged(tag[0]['name'], latest=True)
+    builds = koji.multiCall()
+
+    result = []
+    for i, tag in enumerate(sidetags_info):
+        if (contains_builds and builds[i][0]) or not contains_builds:
+            result.append({
+                'id': tag[0]['id'],
+                'name': tag[0]['name'],
+                'sidetag_user': tag[0]['extra']['sidetag_user'],
+                'builds': builds[i][0]
+            })
+
+    return result
+
+
+def _get_active_updates(request):
+    query = models.Update.query
+
+    update_status = [models.UpdateStatus.pending, models.UpdateStatus.testing]
+    query = query.filter(sa.sql.or_(*[models.Update.status == s for s in update_status]))
+
+    user = models.User.get(request.user.name)
+
+    query = query.filter(models.Update.user == user)
+
+    query = query.order_by(models.Update.date_submitted.desc())
+
+    return query.all()
+
+
+def _get_active_overrides(request):
+    query = models.BuildrootOverride.query
+
+    query = query.filter(models.BuildrootOverride.expired_date.is_(None))
+    user = models.User.get(request.user.name)
+
+    query = query.filter(models.BuildrootOverride.submitter == user)
+
+    query = query.order_by(models.BuildrootOverride.submission_date.desc())
+
+    return query.all()
+
+
+@view_config(route_name='prometheus_metric')
+def get_metrics(request):
+    """
+    Provide the metrics to be consumed by prometheus.
+
+    See the metrics_tween.py for hook to collect metrics for page requests
+
+    Args:
+        request (pyramid.request): The current web request.
+    Returns:
+        str: text in the prometheus metrics format.
+    """
+    registry = REGISTRY
+
+    request.response.content_type = CONTENT_TYPE_LATEST
+    resp = Response(
+        content_type=CONTENT_TYPE_LATEST,
+    )
+    resp.body = generate_latest(registry)
+    return resp
+
+
 @view_config(route_name='home', renderer='home.html')
 def home(request):
     """
@@ -221,7 +255,11 @@ def home(request):
         dict: A Dictionary expressing the values described in the docblock for
             _generate_home_page_stats().
     """
-    return _generate_home_page_stats()
+    data = _generate_home_page_stats()
+    if request.user:
+        data['active_updates'] = _get_active_updates(request)
+        data['active_overrides'] = _get_active_overrides(request)
+    return data
 
 
 @view_config(route_name='new_update', renderer='new_update.html')
@@ -230,7 +268,7 @@ def new_update(request):
     Return the new update form.
 
     Args:
-        request (pyramid.util.Request): The current request.
+        request (pyramid.request.Request): The current request.
     Returns:
         dict: A dictionary with four keys. "update" indexes None. "types" indexes a list of the
             possible UpdateTypes. "severities" indexes a list of the possible severity values.
@@ -242,14 +280,13 @@ def new_update(request):
     if not user:
         raise HTTPForbidden("You must be logged in.")
     suggestions = list(models.UpdateSuggestion.values())
-    if six.PY2:  # pragma: no cover
-        suggestions = reversed(suggestions)
     return dict(
         update=None,
         types=reversed(list(models.UpdateType.values())),
         severities=sorted(list(models.UpdateSeverity.values()),
                           key=bodhi.server.util.sort_severity),
         suggestions=suggestions,
+        sidetags=_get_sidetags(request.koji, user=user, contains_builds=True)
     )
 
 
@@ -263,16 +300,18 @@ def latest_candidates(request):
     "package" parameter.
 
     Args:
-        request (pyramid.util.Request): The current request. The package name is specified in the
+        request (pyramid.request.Request): The current request. The package name is specified in the
             request's "package" parameter.
     Returns:
-        list: A list of dictionaries of the found builds. Each dictionary has two keys: "nvr" maps
-            to the build's nvr field, and "id" maps to the build's id.
+        list: A list of dictionaries of the found builds. Each dictionary has 5 keys: "nvr" maps
+            to the build's nvr field, "id" maps to the build's id, "tag_name" is the tag of the
+            build, owner_name is the person who built the package in koji, and 'release_name' is
+            the bodhi release name of the package.
     """
     koji = request.koji
     db = request.db
 
-    def work(pkg, testing):
+    def work(testing, hide_existing, pkg=None, prefix=None):
         result = []
         koji.multicall = True
 
@@ -280,40 +319,82 @@ def latest_candidates(request):
                      .filter(
                          models.Release.state.in_(
                              (models.ReleaseState.pending,
+                              models.ReleaseState.frozen,
                               models.ReleaseState.current)))
 
-        kwargs = dict(package=pkg, latest=True)
+        if hide_existing:
+            # We want to filter out builds associated with an update.
+            # Since the candidate_tag is removed when an update is pushed to
+            # stable, we only need a list of builds that are associated to
+            # updates still in pending state.
+
+            # Don't filter by releases here, because the associated update
+            # might be archived but the build might be inherited into an active
+            # release. If this gives performance troubles later on, caching
+            # this set should be easy enough.
+            associated_build_nvrs = set(
+                row[0] for row in
+                db.query(models.Build.nvr).
+                join(models.Update).
+                filter(models.Update.status == models.UpdateStatus.pending)
+            )
+
+        kwargs = dict(package=pkg, prefix=prefix, latest=True)
+        tag_release = dict()
         for release in releases:
+            tag_release[release.candidate_tag] = release.long_name
+            tag_release[release.testing_tag] = release.long_name
+            tag_release[release.pending_testing_tag] = release.long_name
+            tag_release[release.pending_signing_tag] = release.long_name
             koji.listTagged(release.candidate_tag, **kwargs)
             if testing:
                 koji.listTagged(release.testing_tag, **kwargs)
                 koji.listTagged(release.pending_testing_tag, **kwargs)
-                koji.listTagged(release.pending_signing_tag, **kwargs)
+                if release.pending_signing_tag:
+                    koji.listTagged(release.pending_signing_tag, **kwargs)
 
         response = koji.multiCall() or []  # Protect against None
-
         for taglist in response:
-            for build in taglist[0]:
-                item = {
-                    'nvr': build['nvr'],
-                    'id': build['id'],
-                }
-                # Prune duplicates
-                # https://github.com/fedora-infra/bodhi/issues/450
-                if item not in result:
-                    result.append(item)
+            # if the call to koji results in errors, it returns them
+            # in the reponse as dicts. Here we detect these, and log
+            # the errors
+            if isinstance(taglist, dict):
+                log.error('latest_candidates endpoint asked Koji about a non-existent tag:')
+                log.error(taglist)
+            else:
+                for build in taglist[0]:
+                    if hide_existing and build['nvr'] in associated_build_nvrs:
+                        continue
+
+                    item = {
+                        'nvr': build['nvr'],
+                        'id': build['id'],
+                        'package_name': build['package_name'],
+                        'owner_name': build['owner_name'],
+                    }
+
+                    # The build's tag might not be present in tag_release
+                    # because its associated release is archived and therefore
+                    # filtered out in the query above.
+                    if build['tag_name'] in tag_release:
+                        item['release_name'] = tag_release[build['tag_name']]
+
+                    # Prune duplicates
+                    # https://github.com/fedora-infra/bodhi/issues/450
+                    if item not in result:
+                        result.append(item)
         return result
 
     pkg = request.params.get('package')
+    prefix = request.params.get('prefix')
     testing = asbool(request.params.get('testing'))
-    log.debug('latest_candidate(%r, %r)' % (pkg, testing))
+    hide_existing = asbool(request.params.get('hide_existing'))
+    log.debug('latest_candidate(%r, %r, %r)' % (pkg, testing, hide_existing))
 
-    if not pkg:
-        return []
-
-    result = work(pkg, testing)
-
-    log.debug(result)
+    if pkg:
+        result = work(testing, hide_existing, pkg=pkg)
+    else:
+        result = work(testing, hide_existing, prefix=prefix)
     return result
 
 
@@ -323,7 +404,7 @@ def latest_builds(request):
     Return a list of the latest builds for a given package.
 
     Args:
-        request (pyramid.util.Request): The current request. The request's "package" parameter is
+        request (pyramid.request.Request): The current request. The request's "package" parameter is
             used to pass the package name being queried.
     Returns:
         dict: A dictionary of the release dist tag to the latest build.
@@ -331,7 +412,7 @@ def latest_builds(request):
     builds = {}
     koji = request.koji
     package = request.params.get('package')
-    for tag_type, tags in six.iteritems(models.Release.get_tags(request.db)[0]):
+    for tag_type, tags in models.Release.get_tags(request.db)[0].items():
         for tag in tags:
             try:
                 for build in koji.getLatestBuilds(tag, package=package):
@@ -341,17 +422,43 @@ def latest_builds(request):
     return builds
 
 
-@view_config(route_name='masher_status', renderer='masher.html')
-def masher_status(request):
+@view_config(route_name='get_sidetags', renderer='json')
+def get_sidetags(request):
     """
-    Return the masher status page.
+    Return a list of koji sidetags based on query arguments.
 
     Args:
-        request (pyramid.util.Request): The current request. Unused.
+        request (pyramid.request.Request): The current request. The request's "user" parameter is
+            used to pass the username being queried.
     Returns:
-        dict: An empty dictionary.
+        dict: A list of the sidetags information.
     """
-    return dict()
+    koji = request.koji
+    # 'user': a FAS username, used to only return sidetags from that user
+    user = request.params.get('user')
+    # 'contains_builds': a boolean to only return sidetags with that contain builds
+    contains_builds = asbool(request.params.get('contains_builds'))
+
+    return _get_sidetags(koji, user=user, contains_builds=contains_builds)
+
+
+@view_config(route_name='latest_builds_in_tag', renderer='json')
+def latest_builds_in_tag(request):
+    """
+    Return a list of the latest builds for a given tag.
+
+    Args:
+        request (pyramid.request.Request): The current request. The request's "tag" parameter is
+            used to pass the koji tagname being queried.
+    Returns:
+        dict: A dictionary of the release dist tag to the latest build.
+    """
+    koji = request.koji
+    tag = request.params.get('tag')
+    if not tag:
+        raise HTTPBadRequest("tag parameter is required")
+    else:
+        return koji.listTagged(tag, latest=True)
 
 
 @view_config(route_name='new_override', renderer='override.html')
@@ -360,7 +467,7 @@ def new_override(request):
     Return the new buildroot override form.
 
     Args:
-        request (pyramid.util.Request): The current request.
+        request (pyramid.request.Request): The current request.
     Returns:
         dict: A dictionary of the form {nvr: nvr}, where the request nvr field indexes itself.
     Raises:
@@ -373,46 +480,49 @@ def new_override(request):
     return dict(nvr=nvr)
 
 
-@view_config(route_name='popup_toggle', request_method='POST')
-def popup_toggle(request):
-    """
-    Toggle whether the user is configured to receive fedmsg popups in the web UI.
-
-    Args:
-        request (pyramid.util.Request): The current request.
-    Returns:
-        pyramid.httpexceptions.HTTPFound: A redirect to the "next" field of the request, or the home
-            page if "next" is not defined.
-    Raises:
-        pyramid.exceptions.HTTPForbidden: If the user is not logged in.
-    """
-    # Get the user
-    userid = request.authenticated_userid
-    if userid is None:
-        raise HTTPForbidden("You must be logged in.")
-    user = request.db.query(models.User).filter_by(name=six.text_type(userid)).first()
-
-    # Toggle the value.
-    user.show_popups = not user.show_popups
-
-    # And send the user back
-    return_to = request.params.get('next', request.route_url('home'))
-    return HTTPFound(location=return_to)
-
-
 @view_config(route_name='api_version', renderer='json')
 def api_version(request):
     """
     Return the Bodhi API version.
 
     Args:
-        request (pyramid.util.Request): The current request.
+        request (pyramid.request.Request): The current request.
     Returns:
         dict: A dictionary with a "version" key indexing a string of the Bodhi version.
     """
     return dict(version=bodhi.server.util.version())
 
 
+@view_config(route_name='liveness', renderer='json')
+def liveness(request):
+    """
+    Return 'ok' as a sign of being alive.
+
+    Args:
+        request (pyramid.request.Request): The current request.
+    Returns:
+        str: 'ok'
+    """
+    return 'ok'
+
+
+@view_config(route_name='readyness', renderer='json')
+def readyness(request):
+    """
+    Return 200 if the app can query the db, to signify being ready.
+
+    Args:
+        request (pyramid.request.Request): The current request.
+    Returns:
+        dict: A dictionary with list of services checked.
+    """
+    try:
+        request.db.execute("SELECT 1")
+        return dict(db_session=True)
+    except Exception:
+        raise Exception("App not ready, is unable to execute a trivial select.")
+
+
 @notfound_view_config(append_slash=True)
 def notfound_view(context, request):
     """
@@ -423,7 +533,7 @@ def notfound_view(context, request):
 
     Args:
         context (pyramid.httpexceptions.HTTPNotFound): The 404 not found error.
-        request (pyramid.util.Request): The current request.
+        request (pyramid.request.Request): The current request.
     Returns:
         bodhi.server.services.errors.html_handler: A pyramid.httpexceptions.HTTPError to be rendered
             to the user for the 404.
@@ -432,6 +542,7 @@ def notfound_view(context, request):
 
 
 @view_config(context=HTTPForbidden, accept='text/html')
+@view_config(context=HTTPUnauthorized, accept='text/html')
 @view_config(context=Exception, accept='text/html')
 def exception_html_view(exc, request):
     """
@@ -444,7 +555,7 @@ def exception_html_view(exc, request):
 
     Args:
         exc (Exception): The unhandled exception.
-        request (pyramid.util.Request): The current request.
+        request (pyramid.request.Request): The current request.
     Returns:
         bodhi.server.services.errors.html_handler: A pyramid.httpexceptions.HTTPError to be rendered
             to the user for the given exception.
@@ -452,18 +563,17 @@ def exception_html_view(exc, request):
     errors = getattr(request, 'errors', [])
     status = getattr(exc, 'status_code', 500)
 
-    if status not in (404, 403):
+    if status not in (404, 403, 401):
         log.exception("Error caught.  Handling HTML response.")
     else:
         log.warning(str(exc))
 
     if not len(errors):
-        description = getattr(exc, 'explanation', None) or str(exc)
-
         errors = cornice.errors.Errors(status=status)
-        errors.add('body', description=description)
+        errors.add('body', description=str(exc))
+        request.errors = errors
 
-    return bodhi.server.services.errors.html_handler(errors, request)
+    return bodhi.server.services.errors.html_handler(request)
 
 
 @view_config(context=HTTPForbidden, accept='application/json')
@@ -479,9 +589,9 @@ def exception_json_view(exc, request):
 
     Args:
         exc (Exception): The unhandled exception.
-        request (pyramid.util.Request): The current request.
+        request (pyramid.request.Request): The current request.
     Returns:
-        bodhi.server.services.errors.html_handler: A pyramid.httpexceptions.HTTPError to be rendered
+        bodhi.server.services.errors.json_handler: A pyramid.httpexceptions.HTTPError to be rendered
             to the user for the given exception.
     """
     errors = getattr(request, 'errors', [])
@@ -493,10 +603,17 @@ def exception_json_view(exc, request):
         log.warning(str(exc))
 
     if not len(errors):
-        description = getattr(exc, 'explanation', None) or str(exc)
-
         errors = cornice.errors.Errors(status=status)
-        errors.add('body', description=description, name=exc.__class__.__name__)
+        errors.add('body', description=str(exc), name=exc.__class__.__name__)
         request.errors = errors
 
     return bodhi.server.services.errors.json_handler(request)
+
+
+@view_config(route_name='docs')
+def docs(request):
+    """Legacy: Redirect the previously self-hosted documentation."""
+    major_minor_version = ".".join(METADATA["version"].split(".")[:2])
+    subpath = "/".join(request.subpath)
+    url = f"https://fedora-infra.github.io/bodhi/{major_minor_version}/{subpath}"
+    raise HTTPMovedPermanently(url)
